/*
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.compile;

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.tools.Diagnostic;
import javax.tools.DiagnosticCollector;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.filefilter.FileFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.systemsx.cisd.base.exceptions.CheckedExceptionTunnel;
import ch.systemsx.cisd.base.exceptions.IOExceptionUnchecked;

/**
 * A simple Java compiler. Can compile any number of independent compile units and put them into a
 * jar file.
 * <p>
 * A simple use case with a class' java code in a string is:
 * 
 * <pre>
 *     String sourceCode = ...;
 *     File jarFile = c.tryCompileAndJar(sourceCode);
 *     if (jarFile == null)
 *     {
 *         System.err.println(c.getErrorMessage());
 *     }
 * </pre>
 * 
 * A simple use case with a hierarchy of files is:
 * 
 * <pre>
 *     File sourceDir = ...;
 *     String jarfileName = ...; 
 *     File jarFile = c.tryCompileAndJar(sourceDir, jarfileName);
 *     if (jarFile == null)
 *     {
 *         System.err.println(c.getErrorMessage());
 *     }
 * </pre>
 * 
 * @author Bernd Rinn
 */
public class SimpleJavaCompiler implements Closeable
{
    private static final Log log = LogFactory.getLog(SimpleJavaCompiler.class);

    private final static Pattern classNamePattern = Pattern.compile(
            "^\\s*public\\s+class\\s+(\\w+)", Pattern.MULTILINE);

    private static final int BUFFER_SIZE = 10240;

    /**
     * Creates a temporary directory with <var>prefix</var>
     * 
     * @param rootDirectoryOrNull The root directory of this temporary directory, or
     *            <code>null</code>, if the default temporary directory should be used.
     * @param prefix The prefix to be used for the temporary directory.
     * @return An empty temporary directory. Delete it when you are done with it.
     */
    private static File createTempDirectory(final File rootDirectoryOrNull, String prefix)
            throws IOExceptionUnchecked
    {
        final File temp;

        try
        {
            temp =
                    File.createTempFile(prefix, Long.toString(System.nanoTime()),
                            rootDirectoryOrNull);
        } catch (IOException ex)
        {
            throw CheckedExceptionTunnel.wrapIfNecessary(ex);
        }

        if (temp.delete() == false)
        {
            throw new IOExceptionUnchecked("Could not delete temp file: " + temp.getAbsolutePath());
        }

        if (temp.mkdir() == false)
        {
            throw new IOExceptionUnchecked("Could not create temp directory: "
                    + temp.getAbsolutePath());
        }

        return (temp);
    }

    // This code is from the internet, see
    // http://www.java2s.com/Code/Java/File-Input-Output/CreateJarfile.htm
    private static boolean createJarArchive(File archiveFile, File root, Iterable<File> tobeJared)
    {
        byte buffer[] = new byte[BUFFER_SIZE];
        FileOutputStream stream = null;
        JarOutputStream out = null;
        final long oldLastChanged = archiveFile.lastModified();
        final File archiveFileInProcess =
                new File(archiveFile.getParentFile(), archiveFile.getName() + ".incomplete");
        try
        {
            // Open archive file
            stream = new FileOutputStream(archiveFileInProcess);
            out = new JarOutputStream(stream, new Manifest());

            for (File file : tobeJared)
            {
                if (file == null || file.exists() == false || file.isDirectory())
                {
                    continue; // Just in case...
                }
                if (log.isDebugEnabled())
                {
                    log.debug("Adding " + file.getName() + " to archive " + archiveFile);
                }

                // Add archive entry
                JarEntry jarAdd = new JarEntry(getRelativePath(root, file));
                jarAdd.setTime(file.lastModified());
                out.putNextEntry(jarAdd);

                // Write file to archive
                FileInputStream in = new FileInputStream(file);
                while (true)
                {
                    int nRead = in.read(buffer, 0, buffer.length);
                    if (nRead <= 0)
                    {
                        break;
                    }
                    out.write(buffer, 0, nRead);
                }
                in.close();
            }

            out.close();
            stream.close();
            return true;
        } catch (Exception ex)
        {
            IOUtils.closeQuietly(stream);
            IOUtils.closeQuietly(out);
            archiveFile.delete();
            log.error("Error when creating the jar archive " + archiveFile, ex);
            return false;
        } finally
        {
            // This is a way to communicate to the JRE ClassLoader that the jar archive should be
            // considered changed.
            if (oldLastChanged == archiveFileInProcess.lastModified())
            {
                // Implementation note: 1000ms = 1s is the granularity of what modern file systems
                // save.
                archiveFileInProcess.setLastModified(oldLastChanged + 1000L);
            }
            archiveFile.delete();
            if (archiveFile.exists())
            {
                throw new IOExceptionUnchecked("Cannot delete old archive file "
                        + archiveFile.getAbsolutePath());
            }
            archiveFileInProcess.renameTo(archiveFile);
        }
    }

    private static String getRelativePath(File root, File file)
    {
        final String absoluteRoot = root.getAbsolutePath();
        final String absolutePath = file.getAbsolutePath();
        return absolutePath.startsWith(absoluteRoot) ? absolutePath
                .substring(absoluteRoot.length() + 1) : absolutePath;
    }

    static String tryFindClassname(String classNameOrNull, String classSource)
    {
        if (classNameOrNull != null)
        {
            return classNameOrNull;
        }
        final Matcher matcher = classNamePattern.matcher(classSource);
        if (matcher.find())
        {
            return matcher.group(1);
        } else
        {
            return null;
        }
    }

    private final File tempDir;

    private final JavaCompiler compiler;

    private final List<String> compilerOptions;

    private final StandardJavaFileManager fileManager;

    private final DiagnosticCollector<JavaFileObject> diagnostics;

    private final StringWriter logWriter;

    private String lastClassName;

    private boolean done;

    public SimpleJavaCompiler() throws IOExceptionUnchecked
    {
        this(null);
    }

    public SimpleJavaCompiler(File tempRootDirectoryOrNull) throws IOExceptionUnchecked
    {
        this.compiler = ToolProvider.getSystemJavaCompiler();
        this.fileManager = compiler.getStandardFileManager(null, null, null);
        this.tempDir = createTempDirectory(tempRootDirectoryOrNull, "javaCompile");
        this.diagnostics = new DiagnosticCollector<JavaFileObject>();
        this.logWriter = new StringWriter();
        this.compilerOptions = new ArrayList<String>();
        // set compiler's classpath to be same as the runtime's
        compilerOptions.addAll(Arrays.asList("-classpath", System.getProperty("java.class.path")));
        // set output directory to be the temporary directory
        compilerOptions.addAll(Arrays.asList("-d", tempDir.getAbsolutePath()));

    }

    public String getLastClassName()
    {
        return lastClassName;
    }

    public List<String> getCompilerOptions()
    {
        return compilerOptions;
    }

    public List<Diagnostic<? extends JavaFileObject>> getDiagnostics()
    {
        return diagnostics.getDiagnostics();
    }

    public String getErrorMessage()
    {
        final StringBuilder builder = new StringBuilder();
        for (Diagnostic<? extends JavaFileObject> d : getDiagnostics())
        {
            builder.append(d.toString());
            builder.append('\n');
        }
        builder.append(logWriter.toString());
        return builder.toString();
    }

    private void checkDone()
    {
        if (done)
        {
            throw new IllegalStateException(
                    "Cannot use SImpleJavaCompiler object after cleanUp() has been called.");
        }
    }

    public JavaFileObject tryGetFileObjectFromSourceCode(String classSource)
    {
        return tryGetFileObjectFromSourceCode(null, classSource);
    }

    public JavaFileObject tryGetFileObjectFromSourceCode(String classNameOrNull, String classSource)
    {
        checkDone();
        final String className = tryFindClassname(classNameOrNull, classSource);
        if (className == null)
        {
            if (log.isWarnEnabled())
            {
                log.warn("Couldn't determine classname for source code:\n" + classSource);
            }
            return null;
        }
        return new JavaSourceFromString(className, classSource);
    }

    public JavaFileObject tryGetFileObjectForFile(File javaFile)
    {
        checkDone();
        final Iterator<? extends JavaFileObject> javaFileObjects =
                fileManager.getJavaFileObjects(javaFile).iterator();
        return javaFileObjects.hasNext() ? javaFileObjects.next() : null;
    }

    public Iterable<? extends JavaFileObject> getFileObjectsForDirectory(File directory)
    {
        checkDone();
        @SuppressWarnings("unchecked")
        final Collection<File> javaFiles = FileUtils.listFiles(directory, new String[]
            { "java" }, true);
        return fileManager.getJavaFileObjectsFromFiles(javaFiles);
    }

    public boolean compile(Iterable<? extends JavaFileObject> fileObjects)
            throws IOExceptionUnchecked
    {
        checkDone();
        final Boolean okObj =
                compiler.getTask(logWriter, fileManager, diagnostics, compilerOptions, null,
                        fileObjects).call();
        final boolean ok = okObj != null && okObj.booleanValue();
        if (ok)
        {
            JavaFileObject lastFile = getLast(fileObjects);
            if (lastFile != null)
            {
                lastClassName = lastFile.getName();
                if (lastClassName.endsWith(".java"))
                {
                    lastClassName =
                            lastClassName.substring(0, lastClassName.length() - ".java".length());
                }
            }
        }
        return ok;
    }

    private JavaFileObject getLast(Iterable<? extends JavaFileObject> fileObjects)
    {
        if (fileObjects instanceof List)
        {
            @SuppressWarnings("unchecked")
            final List<JavaFileObject> fileObjectList = (List<JavaFileObject>) fileObjects;
            return fileObjectList.get(fileObjectList.size() - 1);
        } else
        {
            JavaFileObject lastFile = null;
            Iterator<? extends JavaFileObject> it = fileObjects.iterator();
            while (it.hasNext())
            {
                lastFile = it.next();
            }
            return lastFile;
        }
    }

    public boolean compile(JavaFileObject... fileObjects) throws IOExceptionUnchecked
    {
        return compile(Arrays.asList(fileObjects));
    }

    public void addFile(File source, String destPath) throws IOExceptionUnchecked
    {
        final File destFile = new File(tempDir, destPath);
        destFile.getParentFile().mkdirs();
        try
        {
            FileUtils.copyFile(source, destFile);
        } catch (IOException ex)
        {
            throw CheckedExceptionTunnel.wrapIfNecessary(ex);
        }
    }

    public void addContent(InputStream istream, String destPath) throws IOExceptionUnchecked
    {
        final File destFile = new File(tempDir, destPath);
        destFile.getParentFile().mkdirs();
        OutputStream ostream = null;
        try
        {
            ostream = new FileOutputStream(destFile);
            IOUtils.copyLarge(istream, ostream);
        } catch (IOException ex)
        {
            throw CheckedExceptionTunnel.wrapIfNecessary(ex);
        } finally
        {
            IOUtils.closeQuietly(ostream);
        }
    }

    public File tryCreateJar()
    {
        return tryCreateJar(null, null);
    }

    public File tryCreateJar(final File outputDirectoryOrNull)
    {
        return tryCreateJar(outputDirectoryOrNull, null);
    }

    public File tryCreateJar(final String jarFilenameOrNull)
    {
        return tryCreateJar(null, jarFilenameOrNull);
    }

    public File tryCreateJar(final File outputDirectoryOrNull, final String jarFilenameOrNull)
    {
        checkDone();
        final String jarFilename = (jarFilenameOrNull != null) ? jarFilenameOrNull : lastClassName;
        if (jarFilename == null)
        {
            if (logWriter.toString().isEmpty() == false)
            {
                logWriter.append('\n');
            }
            logWriter
                    .append("Cannot find name for jar file (neither explicity given nor inferreable).");
            return null;
        }

        final File outputJarfile =
                (outputDirectoryOrNull != null) ? new File(outputDirectoryOrNull, jarFilename
                        + ".jar") : new File(jarFilename + ".jar");
        @SuppressWarnings("unchecked")
        final Collection<File> classFiles =
                FileUtils.listFiles(tempDir, FileFileFilter.FILE, TrueFileFilter.TRUE);
        if (createJarArchive(outputJarfile, tempDir, classFiles))
        {
            return outputJarfile;
        } else
        {
            return null;
        }
    }

    public File tryCompileAndJarFromString(String classSource, File outputDir)
            throws IOExceptionUnchecked
    {
        checkDone();
        try
        {
            JavaFileObject fileObject = tryGetFileObjectFromSourceCode(classSource);
            if (fileObject == null)
            {
                return null;
            }
            if (compile(fileObject) == false)
            {
                return null;
            }
            return tryCreateJar(outputDir);
        } finally
        {
            close();
        }
    }

    public File tryCompileAndJarFromFiles(File sourceDirectory, File outputDir, String jarFilename)
            throws IOExceptionUnchecked
    {
        try
        {
            Iterable<? extends JavaFileObject> fileObjects =
                    getFileObjectsForDirectory(sourceDirectory);
            if (compile(fileObjects) == false)
            {
                return null;
            }
            return tryCreateJar(outputDir);
        } finally
        {
            close();
        }
    }

    public File getTempDir()
    {
        return tempDir;
    }

    public StringWriter getLogWriter()
    {
        return logWriter;
    }

    @Override
    public void close() throws IOExceptionUnchecked
    {
        try
        {
            FileUtils.deleteDirectory(tempDir);
        } catch (IOException ex)
        {
            throw CheckedExceptionTunnel.wrapIfNecessary(ex);
        }
        done = true;
    }

}
