/*
 * Copyright 2005-2009 Lukas Blunschi
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.hotdeploy;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import org.apache.commons.io.IOUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.cisd.hotdeploy.PluginEvent.PluginEventType;

/**
 * The <code>PluginContainer</code> manages a map of registered plugins. It supports retrieval of
 * plugins by classname, type and plugin archive.
 * <p>
 * There can be multiple instances of this class which are in charge of monitoring different
 * directories. <b>Do not try to let multiple instances monitor the same directory!</b>
 * 
 * @author Lukas Blunschi
 * @author Bernd Rinn
 */
// For the PluginLoader it allows to register new plugins, update change plugins and unregister
// removed plugins.
public final class PluginContainer
{
    private static final Log log = LogFactory.getLog(PluginContainer.class);

    /**
     * The maximal time (in s) to wait for the loader to finish a cycle.
     */
    private static final long MAX_TIME_IN_SECS_TO_WAIT_FOR_LOAD_CYCLE = 5L;

    private static final String MAIN_PLUGIN_CONTAINER_NAME = "MAIN";

    /** The map of instances of this class. */
    private static final Map<String, PluginContainer> instances = Collections
            .synchronizedMap(new HashMap<String, PluginContainer>());

    /**
     * The name of this instance.
     */
    private final String containerName;

    /** <code>true</code> if we should run GC on shutdown. */
    private final boolean runGcOnShutdown;

    /** Reference to the loader thread. */
    private Thread loaderThread;

    /**
     * Reference to the loader.
     */
    private PluginLoader loader;

    /**
     * The notifier for {@link PuginEvent}s.
     */
    private PluginEventNotifier notifier;

    // Thread-safety information:
    // All methods reading or writing classnameMap or pluginDirToPluginSetMap <i>must</i> hold
    // the lock on the monitor of this plugin container instance.

    /** Maps from plugin classnames to plugin descriptors. */
    private Map<String, PluginDescriptor> classNameMap = new HashMap<String, PluginDescriptor>();

    /** Maps from plugin names to plugin descriptors. */
    private Map<String, PluginDescriptor> pluginNameMap = new HashMap<String, PluginDescriptor>();

    /** Maps from plugin jar files to plugin descriptors. */
    private Map<File, PluginDescriptor> pluginJarMap = new HashMap<File, PluginDescriptor>();

    /**
     * A map from plugin directory to a set of plugin classnames.
     */
    private Map<File, Set<PluginDescriptor>> pluginDirToPluginSetMap =
            new HashMap<File, Set<PluginDescriptor>>();

    /**
     * Private constructor. Used only by the initHotDeployment function. Initializes and starts the
     * PluginLoader thread.
     * 
     * @param params
     */
    private PluginContainer(String name, PluginContainerParameters params)
    {
        this.containerName = name;
        long begin = System.nanoTime();

        // Set GC on shutdown parameter
        runGcOnShutdown = params.isRunGCOnShutdown();
        if (runGcOnShutdown)
        {
            log.info("Running GC on shutdown.");
        } else
        {
            log.info("Not running GC on shutdown.");
        }
        notifier = new PluginEventNotifier();

        // start the plugin loader thread
        loader = new PluginLoader(this, params);
        loaderThread = new Thread(loader, "ch.ethz.cisd.hotdeploy.PluginLoader (" + name + ")");
        loaderThread.setDaemon(true);
        final CountDownLatch firstRoundSync = loader.getCycleFinishedBarrier();
        loaderThread.start();

        // wait for plugin loader to finish first round
        try
        {
            firstRoundSync.await(MAX_TIME_IN_SECS_TO_WAIT_FOR_LOAD_CYCLE, TimeUnit.SECONDS);
        } catch (InterruptedException ex)
        {
        }

        String ms = String.valueOf((System.nanoTime() - begin) / 1E6);
        log.info("PluginContainer initialized (" + ms + "ms).");
        log.info(BuildAndEnvironmentInfo.INSTANCE.getApplicationName() + " v"
                + BuildAndEnvironmentInfo.INSTANCE.getFullVersion() + " is ready.");
    }

    public String getContainerName()
    {
        return containerName;
    }

    /**
     * @param listener
     * @see ch.ethz.cisd.hotdeploy.PluginEventNotifier#addListener(ch.ethz.cisd.hotdeploy.PluginEventListener)
     */
    public void addListener(PluginEventListener listener)
    {
        notifier.addListener(listener);
    }

    /**
     * @param listener
     * @see ch.ethz.cisd.hotdeploy.PluginEventNotifier#removeListener(ch.ethz.cisd.hotdeploy.PluginEventListener)
     */
    public void removeListener(PluginEventListener listener)
    {
        notifier.removeListener(listener);
    }

    /**
     * Returns a set with all currently registered plugin directories.
     * 
     * @return A copy of the set of plugin directories.
     */
    public Set<File> getPluginDirectories()
    {
        return loader.getPluginDirectories();
    }

    /**
     * Add <var>pluginDirectory</var> to the search lists of plugins. Waits for the refresh cycle to
     * be finished.
     */
    public void addPluginDirectory(File pluginDirectory)
    {
        addPluginDirectory(pluginDirectory, true);
    }

    /**
     * Add <var>pluginDirectory</var> to the search lists of plugins.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void addPluginDirectory(File pluginDirectory, boolean synchronous)
    {
        if (pluginDirectory.isDirectory() == false)
        {
            log.warn("path " + pluginDirectory + " is not a directory");
        }
        loader.addPluginDirectory(pluginDirectory);
        log.info("add plugin directory " + pluginDirectory);
        refresh(synchronous);
    }

    /**
     * Add all <var>pluginDirectoriesToBeAdded</var> to the search lists of plugins. Waits for the
     * refresh cycle to be finished.
     */
    public void addPluginDirectories(Collection<File> pluginDirectoriesToBeAdded)
    {
        addPluginDirectories(pluginDirectoriesToBeAdded, true);
    }

    /**
     * Add all <var>pluginDirectoriesToBeAdded</var> to the search lists of plugins.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void addPluginDirectories(Collection<File> pluginDirectoriesToBeAdded,
            boolean synchronous)
    {
        for (File pluginDirectory : pluginDirectoriesToBeAdded)
        {
            if (pluginDirectory.isDirectory() == false)
            {
                log.warn("path " + pluginDirectory + " is not a directory");
            }
        }
        loader.addPluginDirectories(pluginDirectoriesToBeAdded);
        log.info("add plugin directories " + pluginDirectoriesToBeAdded);
        refresh(synchronous);
    }

    /**
     * Remove <var>pluginDirectory</var> from the search lists of plugins. Waits for the refresh
     * cycle to be finished.
     */
    public void removePluginDirectory(File pluginDirectory)
    {
        removePluginDirectory(pluginDirectory, true);
    }

    /**
     * Remove <var>pluginDirectory</var> from the search lists of plugins.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void removePluginDirectory(File pluginDirectory, boolean synchronous)
    {
        loader.removePluginDirectory(pluginDirectory);
        log.info("remove plugin directory " + pluginDirectory);
        refresh(synchronous);
    }

    /**
     * Remove all <var>pluginDirectoriesToBeRemoved</var> from the search lists of plugins. Waits
     * for the refresh cycle to be finished.
     */
    public void removePluginDirectories(Collection<File> pluginDirectoriesToBeRemoved)
    {
        removePluginDirectories(pluginDirectoriesToBeRemoved, true);
    }

    /**
     * Remove all <var>pluginDirectoriesToBeRemoved</var> from the search lists of plugins.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void removePluginDirectories(Collection<File> pluginDirectoriesToBeRemoved,
            boolean synchronous)
    {
        loader.removePluginDirectories(pluginDirectoriesToBeRemoved);
        log.info("remove plugin directories " + pluginDirectoriesToBeRemoved);
        refresh(synchronous);
    }

    /**
     * Sets the <var>pluginDirectories</var> as the search lists of plugins, removing all plugin
     * directories set previously. Wait for the refresh cycle to be finished.
     */
    public void setPluginDirectories(Collection<File> newPluginDirectories)
    {
        setPluginDirectories(newPluginDirectories, true);
    }

    /**
     * Sets the <var>pluginDirectories</var> as the search lists of plugins, removing all plugin
     * directories set previously.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void setPluginDirectories(Collection<File> newPluginDirectories, boolean synchronous)
    {
        for (File pluginDirectory : newPluginDirectories)
        {
            if (pluginDirectory.isDirectory() == false)
            {
                log.warn("path " + pluginDirectory + " is not a directory");
            }
        }
        loader.setPluginDirectories(newPluginDirectories);
        log.info("set plugin directories to " + newPluginDirectories);
        refresh(synchronous);
    }

    /**
     * Register a new plugin or updates an already registered plugin.
     * 
     * @param desc the plugin deployment descriptor to register if its null, to nothing
     * @param newPlugin <code>true</code>, if this is a new plugin, <code>false</code>, if it is the
     *            update of an already registered plugin.
     */
    synchronized void register(PluginDescriptor desc, boolean newPlugin)
    {
        if (desc == null)
        {
            // protect against programming error
            String msg = "register plugin called with null argument.";
            log.error(msg, new Exception(msg));
            return;
        }
        String classname = desc.getPluginClassname();
        // add mapping
        classNameMap.put(classname, desc);
        pluginNameMap.put(desc.getPluginName(), desc);
        pluginJarMap.put(desc.getPluginArchive(), desc);
        addArchiveToPluginDirMap(desc);

        // notify listeners
        final PluginEventType eventType =
                newPlugin ? PluginEventType.REGISTER_NEW_PLUGIN : PluginEventType.UPDATE_PLUGIN;
        notifier.notifyListeners(new PluginEvent(eventType, desc.getPluginClass(), desc
                .getPluginArchive()));
    }

    private void addArchiveToPluginDirMap(PluginDescriptor pluginDesc)
    {
        final File pluginDir = pluginDesc.getPluginDirectory();
        Set<PluginDescriptor> pluginsInDir = pluginDirToPluginSetMap.get(pluginDir);
        if (pluginsInDir == null)
        {
            pluginsInDir = new HashSet<PluginDescriptor>();
            pluginDirToPluginSetMap.put(pluginDir, pluginsInDir);
        }
        pluginsInDir.add(pluginDesc);
    }

    /**
     * Unregister a plugin.
     * <p>
     * This function is idempotent.
     * <p>
     * It will inform observers of the change so they have a chance to release references.
     * 
     * @param desc the plugin to unregister if null, does nothing.
     */
    synchronized void unregister(PluginDescriptor desc)
    {
        if (desc == null)
        {
            // protect against programming error
            String msg = "unregister plugin called with null argument.";
            log.error(msg, new Exception(msg));
            return;
        }
        final boolean isUnregistered = (classNameMap.remove(desc.getPluginClassname()) != null);
        if (isUnregistered)
        {
            pluginNameMap.remove(desc.getPluginName());
            removePluginFromPluginDirMap(desc);
            pluginJarMap.remove(desc.getPluginArchive());
            // Notify listeners. Listeners are expected to release all references to this plugin.
            notifier.notifyListeners(new PluginEvent(PluginEventType.UNREGISTER_PLUGIN, desc
                    .getPluginClass(), desc.getPluginArchive()));
        }
    }

    private void removePluginFromPluginDirMap(PluginDescriptor pluginDesc)
    {
        final Set<PluginDescriptor> pluginsInDir =
                pluginDirToPluginSetMap.get(pluginDesc.getPluginDirectory());
        if (pluginsInDir != null)
        {
            pluginsInDir.remove(pluginDesc);
        }
    }

    /**
     * Initialize plugin hot deployment. This function MUST be called before a call to
     * {@link #tryGetInstance()}. It starts the <code>PluginLoader</code> thread.
     * 
     * @param params configuration parameters for the PluginContainer and the PluginLoader
     */
    public static PluginContainer initHotDeployment(PluginContainerParameters params)
    {
        return initHotDeployment(MAIN_PLUGIN_CONTAINER_NAME, params);
    }

    /**
     * Initialize plugin hot deployment. This function MUST be called before a call to
     * {@link #tryGetInstance()}. It starts the <code>PluginLoader</code> thread.
     * 
     * @param params configuration parameters for the PluginContainer and the PluginLoader
     */
    public static PluginContainer initHotDeployment(String containerName,
            PluginContainerParameters params)
    {
        PluginContainer instance = instances.get(containerName);
        if (instance == null)
        {
            instance = new PluginContainer(containerName, params);
            instances.put(containerName, instance);
        }
        return instance;
    }

    /**
     * Initialize plugin hot deployment. This function MUST be called before a call to
     * {@link #tryGetInstance()}. It starts the <code>PluginLoader</code> thread.
     * 
     * @param containerName The name of the container to initialize.
     * @param scanIntervalMillis The scan interval in milliseconds. Use a negative value to not
     *            re-scan periodically.
     */
    public static PluginContainer initHotDeployment(String containerName, long scanIntervalMillis)
    {
        return initHotDeployment(containerName,
                new PluginContainerParameters().scanIntervalMillis(scanIntervalMillis));
    }

    /**
     * Initialize plugin hot deployment. This function MUST be called before a call to
     * {@link #tryGetInstance()}. It starts the <code>PluginLoader</code> thread.
     * 
     * @param scanIntervalMillis The scan interval in milliseconds. Use a negative value to not
     *            re-scan periodically.
     */
    public static PluginContainer initHotDeployment(long scanIntervalMillis)
    {
        return initHotDeployment(MAIN_PLUGIN_CONTAINER_NAME, scanIntervalMillis);
    }

    /**
     * Initialize plugin hot deployment with default parameters. This function MUST be called before
     * a call to {@link #tryGetInstance()}. It starts the <code>PluginLoader</code> thread.
     * <p>
     * Note that it has not yet any plugin directory set so won't know of any plugins. Call
     * {@link #addPluginDirectory(File)} to rectify this.
     * 
     * @param containerName The name of the container to initialize.
     */
    public static PluginContainer initHotDeployment(String containerName)
    {
        return initHotDeployment(containerName, new PluginContainerParameters());
    }

    /**
     * Initialize plugin hot deployment with default parameters. This function MUST be called before
     * a call to {@link #tryGetInstance()}. It starts the <code>PluginLoader</code> thread.
     * <p>
     * Note that it has not yet any plugin directory set so won't know of any plugins. Call
     * {@link #addPluginDirectory(File)} to rectify this.
     */
    public static PluginContainer initHotDeployment()
    {
        return initHotDeployment(MAIN_PLUGIN_CONTAINER_NAME);
    }

    /**
     * Return the main instance.
     * 
     * @return The instance or null if initHotDeployment was not called before for the main
     *         instance.
     * @see #initHotDeployment
     */
    public static PluginContainer tryGetInstance()
    {
        return instances.get(MAIN_PLUGIN_CONTAINER_NAME);
    }

    /**
     * Returns the instance with the given <var>containerName</var>.
     * 
     * @param containerName The name of this container.
     * @return The instance or null if initHotDeployment was not called before.
     * @see #initHotDeployment
     */
    public static PluginContainer tryGetInstance(String containerName)
    {
        return instances.get(containerName);
    }

    /**
     * Refresh plugins now.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void refresh(boolean synchronous)
    {
        if (synchronous)
        {
            final CountDownLatch sync = loader.getCycleFinishedBarrier();
            loaderThread.interrupt();
            try
            {
                sync.await(MAX_TIME_IN_SECS_TO_WAIT_FOR_LOAD_CYCLE, TimeUnit.SECONDS);
            } catch (InterruptedException ex)
            {
            }
        } else
        {
            loaderThread.interrupt();
        }
    }

    /**
     * Get a plugin class by its <var>classname</var>.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @return The plugin class or <code>null</code> if not registered or <var>classname<var> is
     *         <code>null</code>.
     */
    public Class<?> tryGetPluginClassByClassname(String classname)
    {
        return tryGetPluginClassByClassname(classname, Object.class);
    }

    /**
     * Get a plugin class by its <var>classname</var>.
     * 
     * @param name The name of the plugin.
     * @return The plugin class or <code>null</code> if not registered or <var>classname<var> is
     *         <code>null</code>.
     */
    public Class<?> tryGetPluginClass(String name)
    {
        return tryGetPluginClass(name, false, Object.class);
    }

    /**
     * Get a plugin by its <var>jarfile</var>.
     * 
     * @param jarfile The plugin jar file of the plugin.
     * @param clazz The class that this plugin should be cast to.
     * @return The plugin class or <code>null</code> if not registered or <var>classname<var> is
     *         <code>null</code>.
     * @throws ClassCastException If the plugin cannot be cast to <var>clazz</var>.
     */
    @SuppressWarnings("unchecked")
    public <T> Class<T> tryGetPluginClass(File jarfile, Class<T> clazz)
    {
        if (jarfile == null)
        {
            return null;
        }
        final PluginDescriptor desc = tryGetPluginDesc(jarfile, true);
        if (desc == null)
        {
            return null;
        }

        checkImplements(desc, clazz);
        return (Class<T>) desc.getPluginClass();
    }

    /**
     * Get a plugin by its <var>jarfile</var>.
     * 
     * @param jarfile The jar file of the plugin.
     * @return The plugin or <code>null</code> if not registered or <var>classname<var> is
     *         <code>null</code>.
     */
    public Class<?> tryGetPluginClass(File jarfile)
    {
        return tryGetPluginClass(jarfile, Object.class);
    }

    /**
     * Get plugin information for a plugin by its <var>classname</var>.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public PluginInfo tryGetPluginInfoByClassname(String classname)
    {
        final PluginDescriptor desc = tryGetPluginDesc(classname, true, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginInfo();
    }

    /**
     * Get plugin information for a plugin by its <var>classname</var>.
     * 
     * @param name The name of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, name is
     *         <code>null</code>, or no plugin information is defined for the plugin.
     */
    public PluginInfo tryGetPluginInfo(String name)
    {
        final PluginDescriptor desc = tryGetPluginDesc(name, false, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginInfo();
    }

    /**
     * Get plugin information for a plugin by its <var>jarfile</var>.
     * 
     * @param jarfile The plugin jar file of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, name is
     *         <code>null</code>, or no plugin information is defined for the plugin.
     */
    public PluginInfo tryGetPluginInfo(File jarfile)
    {
        final PluginDescriptor desc = tryGetPluginDesc(jarfile, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginInfo();
    }

    /**
     * Get the plugin name of the plugin by its <var>classname</var>. If the plugin has a
     * {@link PluginInfo}, it will be {@link PluginInfo#name()}, otherwise it will be the
     * <var>classname</var> itself.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public String tryGetPluginNameByClassname(String classname)
    {
        final PluginDescriptor desc = tryGetPluginDesc(classname, true, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginName();
    }

    /**
     * Get the plugin name of the plugin by its <var>jarfile</var>. If the plugin has a
     * {@link PluginInfo}, it will be {@link PluginInfo#name()}, otherwise it will be the
     * <var>classname</var> itself.
     * 
     * @param jarfile The plugin jar file to get the plugin for.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public String tryGetPluginName(File jarfile)
    {
        final PluginDescriptor desc = tryGetPluginDesc(jarfile, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginName();
    }

    /**
     * Get the plugin's classname of the plugin by its regular <var>name</var>. Note that if the
     * plugin has no {@link PluginInfo}, the name and the classname are the same.
     * 
     * @param name The name of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public String tryGetPluginClassname(String name)
    {
        final PluginDescriptor desc = tryGetPluginDesc(name, false, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginClassname();
    }

    /**
     * Get the plugin's classname of the plugin by its <var>jarfile</var>. Note that if the plugin
     * has no {@link PluginInfo}, the name and the classname are the same.
     * 
     * @param jarfile The plugin jar file of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public String tryGetPluginClassname(File jarfile)
    {
        final PluginDescriptor desc = tryGetPluginDesc(jarfile, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginClassname();
    }

    /**
     * Get the type of the plugin by its <var>classname</var>. If the plugin has a
     * {@link PluginInfo}, it will be {@link PluginInfo#pluginType()}, otherwise it will be the
     * plugin class itself.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public Class<?> tryGetPluginTypeByClassname(String classname)
    {
        final PluginDescriptor desc = tryGetPluginDesc(classname, true, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginType();
    }

    /**
     * Get the type of the plugin. If the plugin has a {@link PluginInfo}, it will be
     * {@link PluginInfo#pluginType()}, otherwise it will be the plugin class itself.
     * 
     * @param name The name of the plugin.
     * @return the plugin information or <code>null</code> if plugin is not registered, classname is
     *         <code>null</code>, or no plugin information is defined.
     */
    public Class<?> tryGetPluginType(String name)
    {
        final PluginDescriptor desc = tryGetPluginDesc(name, false, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginType();
    }

    /**
     * Returns <code>true</code> if a plugin with given <var>classname</var> exists.
     * 
     * @param classname The fully qualified class name of the plugin.
     */
    public boolean hasPluginByClassname(String classname)
    {
        return tryGetPluginClassByClassname(classname) != null;
    }

    /**
     * Returns <code>true</code> if a plugin with given <var>name</var> exists.
     * 
     * @param name The name of the plugin.
     */
    public boolean hasPlugin(String name)
    {
        return tryGetPluginClass(name) != null;
    }

    /**
     * Returns <code>true</code> if the plugin with <var>classname</var> has a constructor for the
     * <var>constructorArgumentTypes</var>.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @param constructorArgumentTypes The types of the constructor arguments. Empty for default
     *            constructor.
     */
    public boolean hasConstructorByClassname(String classname, Class<?>... constructorArgumentTypes)
    {
        final Class<?> pluginClass = tryGetPluginClassByClassname(classname);
        if (pluginClass == null)
        {
            return false;
        }
        return ReflectionUtils.tryGetConstructor(pluginClass, constructorArgumentTypes) != null;
    }

    /**
     * Returns <code>true</code> if the plugin with <var>name</var> has a constructor for the
     * <var>constructorArgumentTypes</var>.
     * 
     * @param name The name of the plugin.
     * @param constructorArgumentTypes The types of the constructor arguments. Empty for default
     *            constructor.
     */
    public boolean hasConstructor(String name, Class<?>... constructorArgumentTypes)
    {
        final Class<?> pluginClass = tryGetPluginClass(name);
        if (pluginClass == null)
        {
            return false;
        }
        return ReflectionUtils.tryGetConstructor(pluginClass, constructorArgumentTypes) != null;
    }

    /**
     * Returns <code>true</code> if the plugin in <var>jarfile</var> has a constructor for the
     * <var>constructorArgumentTypes</var>.
     * 
     * @param jarfile The plugin jar file of the plugin.
     * @param constructorArgumentTypes The types of the constructor arguments. Empty for default
     *            constructor.
     */
    public boolean hasConstructor(File jarfile, Class<?>... constructorArgumentTypes)
    {
        final Class<?> pluginClass = tryGetPluginClass(jarfile);
        if (pluginClass == null)
        {
            return false;
        }
        return ReflectionUtils.tryGetConstructor(pluginClass, constructorArgumentTypes) != null;
    }

    /**
     * Returns the plugin archive file for the given plugin <var>classname</var>.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @return The plugin archive or <code>null</code>, if classname is <code>null</code> or no
     *         plugin archive can be found for it.
     */
    public File tryGetPluginArchiveByClassname(String classname)
    {
        if (classname == null)
        {
            return null;
        }
        final PluginDescriptor desc = tryGetPluginDesc(classname, true, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginArchive();
    }

    /**
     * Returns the plugin archive file for the given plugin <var>name</var>.
     * 
     * @param name The name of the plugin.
     * @return The plugin archive or <code>null</code>, if name is <code>null</code> or no plugin
     *         archive can be found for it.
     */
    public File tryGetPluginArchive(String name)
    {
        if (name == null)
        {
            return null;
        }
        final PluginDescriptor desc = tryGetPluginDesc(name, false, true);
        if (desc == null)
        {
            return null;
        }
        return desc.getPluginArchive();
    }

    /**
     * Returns a resource from the given plugin archive, or <code>null</code>, if either the archive
     * or the resource could not be found.
     */
    public byte[] tryGetPluginResource(final File pluginArchive, final String resource)
    {
        if (pluginArchive == null)
        {
            return null;
        }
        return loader.tryRunLocked(pluginArchive, new PluginLoader.ClassLoaderCallable<byte[]>()
            {
                @Override
                public byte[] call(ClassLoader classLoader) throws Exception
                {
                    final InputStream istream = classLoader.getResourceAsStream(resource);
                    try
                    {
                        final ByteArrayOutputStream ostream = new ByteArrayOutputStream();
                        IOUtils.copy(istream, ostream);
                        return ostream.toByteArray();
                    } finally
                    {
                        istream.close();
                    }
                }
            });
    }

    /**
     * Get a plugin by its <var>classname</var>.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @param clazz The class that this plugin should be cast to.
     * @return The plugin or <code>null</code> if not registered or <var>classname<var> is
     *         <code>null</code>.
     * @throws ClassCastException If the plugin cannot be cast to <var>clazz</var>.
     */
    public <T> Class<T> tryGetPluginClassByClassname(String classname, Class<T> clazz)
    {
        return tryGetPluginClass(classname, true, clazz);
    }

    /**
     * Get a plugin by its <var>name</var>.
     * 
     * @param name The name of the plugin.
     * @param clazz The class that this plugin should be cast to.
     * @return The plugin or <code>null</code> if not registered or <var>name<var> is
     *         <code>null</code>.
     * @throws ClassCastException If the plugin cannot be cast to <var>clazz</var>.
     */
    public <T> Class<T> tryGetPluginClass(String name, Class<T> clazz)
    {
        return tryGetPluginClass(name, false, clazz);
    }

    /**
     * Get a plugin by its <var>classname</var>.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @param clazz The class that this plugin should be cast to.
     * @return The plugin or <code>null</code> if not registered or <var>classname<var> is
     *         <code>null</code>.
     * @throws ClassCastException If the plugin cannot be cast to <var>clazz</var>.
     */
    @SuppressWarnings("unchecked")
    <T> Class<T> tryGetPluginClass(String classname, boolean useClassname, Class<T> clazz)
    {
        if (classname == null)
        {
            return null;
        }
        final PluginDescriptor desc = tryGetPluginDesc(classname, useClassname, true);
        if (desc == null)
        {
            return null;
        }

        checkImplements(desc, clazz);
        return (Class<T>) desc.getPluginClass();
    }

    private <T> void checkImplements(final PluginDescriptor desc, Class<T> clazz)
    {
        if (desc.isImplementing(clazz) == false)
        {
            throw new ClassCastException("Class " + desc.getPluginClassname() + " is not a "
                    + clazz.getName());
        }
    }

    PluginDescriptor tryGetPluginDesc(String name, boolean useClassname, boolean updateIfOutdated)
    {
        PluginDescriptor desc;
        synchronized (this)
        {
            desc = useClassname ? classNameMap.get(name) : pluginNameMap.get(name);
        }
        if (updateIfOutdated && updateIfNecessary(desc))
        {
            synchronized (this)
            {
                desc = useClassname ? classNameMap.get(name) : pluginNameMap.get(name);
            }
        }
        return desc;
    }

    private boolean updateIfNecessary(PluginDescriptor desc)
    {
        if (desc != null && desc.hasPluginArchiveChanged())
        {
            long start = System.currentTimeMillis();
            refresh(true);
            if (log.isDebugEnabled())
            {
                log.debug("refresh took " + (System.currentTimeMillis() - start) + "ms");
            }
            return true;
        }
        return false;
    }

    PluginDescriptor tryGetPluginDesc(File jarfile, boolean updateIfOutdated)
    {
        PluginDescriptor desc;
        synchronized (this)
        {
            desc = pluginJarMap.get(jarfile);
        }
        if (updateIfOutdated && updateIfNecessary(desc))
        {
            synchronized (this)
            {
                desc = pluginJarMap.get(jarfile);
            }
        }
        return desc;
    }

    /**
     * Creates a new instance of a plugin.
     * 
     * @param classname The fully qualified class name of the plugin.
     * @param workerInterface The worker interface to cast the plugin to.
     * @return The plugin or <code>null</code> if not registered or classname is <code>null</code>.
     * @throw IllegalArgumentException If the class cannot be instantiated, e.g. because it is
     *        abstract or because it is lacking a constructor with the given arguments.
     * @throw ClassCastException If the instance cannot be cast to <var>workerInterface</var>.
     * @throw {@link RuntimeException} or any sub-class of it: rethrown if calling the constructor
     *        fails.
     */
    public <T> T tryCreatePluginByClassname(String classname, Class<T> workerInterface,
            Object... constructorArguments)
    {
        final Class<T> pluginClass = tryGetPluginClassByClassname(classname, workerInterface);
        if (pluginClass == null)
        {
            return null;
        }
        return ReflectionUtils.create(workerInterface, pluginClass, constructorArguments);
    }

    /**
     * Creates a new instance of a plugin.
     * 
     * @param name The name of the plugin.
     * @param workerInterface The worker interface to cast the plugin to.
     * @return The plugin or <code>null</code> if not registered or name is <code>null</code>.
     * @throw IllegalArgumentException If the class cannot be instantiated, e.g. because it is
     *        abstract or because it is lacking a constructor with the given arguments.
     * @throw ClassCastException If the instance cannot be cast to <var>workerInterface</var>.
     * @throw {@link RuntimeException} or any sub-class of it: rethrown if calling the constructor
     *        fails.
     */
    public <T> T tryCreatePlugin(String name, Class<T> workerInterface,
            Object... constructorArguments)
    {
        final Class<T> pluginClass = tryGetPluginClass(name, workerInterface);
        if (pluginClass == null)
        {
            return null;
        }
        return ReflectionUtils.create(workerInterface, pluginClass, constructorArguments);
    }

    /**
     * Creates a new instance of a plugin, given by its <var>jarfile</var>.
     * 
     * @param jarfile The jarfile of the plugin.
     * @param workerInterface The worker interface to cast the plugin to.
     * @return The plugin or <code>null</code> if not registered or name is <code>null</code>.
     * @throw IllegalArgumentException If the class cannot be instantiated, e.g. because it is
     *        abstract or because it is lacking a constructor with the given arguments.
     * @throw ClassCastException If the instance cannot be cast to <var>workerInterface</var>.
     * @throw {@link RuntimeException} or any sub-class of it: rethrown if calling the constructor
     *        fails.
     */
    public <T> T tryCreatePlugin(File jarfile, Class<T> workerInterface,
            Object... constructorArguments)
    {
        final Class<T> pluginClass = tryGetPluginClass(jarfile, workerInterface);
        if (pluginClass == null)
        {
            return null;
        }
        return ReflectionUtils.create(workerInterface, pluginClass, constructorArguments);
    }

    /**
     * Get the fully qualified class names of all plugins registered.
     * 
     * @return A set of all plugin class names registered.
     */
    public synchronized Set<String> getPluginClassnames()
    {
        final Set<String> classnames = new HashSet<String>();
        classnames.addAll(classNameMap.keySet());
        return classnames;
    }

    /**
     * Get the names of all plugins registered.
     * 
     * @return A set of all plugin names registered.
     */
    public synchronized Set<String> getPluginNames()
    {
        final Set<String> names = new HashSet<String>();
        names.addAll(pluginNameMap.keySet());
        return names;
    }

    /**
     * Get the types of all plugins registered.
     * 
     * @return A set of all plugin names registered.
     */
    public synchronized Set<Class<?>> getPluginTypes()
    {
        final Set<Class<?>> types = new HashSet<Class<?>>();
        for (PluginDescriptor desc : classNameMap.values())
        {
            types.add(desc.getPluginType());
        }
        return types;
    }

    /**
     * Get the class names of all plugins registered that implement <var>pluginType</var>.
     * 
     * @return A set of all fully qualified class names of plugins that implement the
     *         <var>pluginType</var>.
     */
    public synchronized Set<String> getPluginClassnames(Class<?> pluginType)
    {
        final Set<String> classnames = new HashSet<String>();
        for (PluginDescriptor desc : classNameMap.values())
        {
            if (pluginType == null || desc.isImplementing(pluginType))
            {
                classnames.add(desc.getPluginClassname());
            }
        }
        return classnames;
    }

    /**
     * Get all plugins registered.
     * 
     * @return A list of descriptors for all plugins currently registered.
     */
    public synchronized List<PluginDescriptor> getPlugins()
    {
        return new ArrayList<PluginDescriptor>(classNameMap.values());
    }

    /**
     * Get all plugins registered that implement <var>pluginType</var>.
     * 
     * @return A list of descriptors for all plugins currently registered.
     */
    public synchronized List<PluginDescriptor> getPlugins(Class<?> pluginType)
    {
        final ArrayList<PluginDescriptor> list =
                new ArrayList<PluginDescriptor>(classNameMap.size());
        for (PluginDescriptor desc : classNameMap.values())
        {
            if (pluginType == null || desc.isImplementing(pluginType))
            {
                list.add(desc);
            }
        }
        return list;
    }

    /**
     * Get the names of all plugins registered that implement <var>pluginType</var>.
     * 
     * @return A set of all plugin names that implement the <var>pluginType</var>.
     */
    public synchronized Set<String> getPluginNames(Class<?> pluginType)
    {
        final Set<String> classnames = new HashSet<String>();
        for (PluginDescriptor desc : classNameMap.values())
        {
            if (pluginType == null || desc.isImplementing(pluginType))
            {
                classnames.add(desc.getPluginName());
            }
        }
        return classnames;
    }

    /**
     * Get the class names of all plugins registered from archives in <var>pluginDir</var>.
     * 
     * @return A set of fully qualified class names of plugins registered from archives in
     *         <var>pluginDir</var>.
     */
    public synchronized Set<String> getPluginClassnamesForArchiveDir(File pluginDir)
    {
        Set<String> classnames = new HashSet<String>();
        Set<PluginDescriptor> pluginsInPluginDir = pluginDirToPluginSetMap.get(pluginDir);
        if (pluginsInPluginDir != null)
        {
            for (PluginDescriptor desc : pluginsInPluginDir)
            {
                classnames.add(desc.getPluginClassname());
            }
        }
        return classnames;
    }

    /**
     * Get the names of all plugins registered from archives in <var>pluginDir</var>.
     * 
     * @return A set of plugin names registered from archives in <var>pluginDir</var>.
     */
    public synchronized Set<String> getPluginNamesForArchiveDir(File pluginDir)
    {
        Set<String> classnames = new HashSet<String>();
        Set<PluginDescriptor> pluginsInPluginDir = pluginDirToPluginSetMap.get(pluginDir);
        if (pluginsInPluginDir != null)
        {
            for (PluginDescriptor desc : pluginsInPluginDir)
            {
                classnames.add(desc.getPluginName());
            }
        }
        return classnames;
    }

    /**
     * Get the class name of the plugin registered from <var>pluginArchive</var>.
     * 
     * @return The fully qualified name of the plugin class registered from
     *         <var>pluginArchive</var>, or <code>null</code>, if no plugin was registered from this
     *         archive.
     */
    public synchronized String getPluginClassnameForArchive(File pluginArchive)
    {
        final Set<PluginDescriptor> pluginsInPluginDir =
                pluginDirToPluginSetMap.get(pluginArchive.getParentFile());
        if (pluginsInPluginDir == null)
        {
            return null;
        }
        for (PluginDescriptor desc : pluginsInPluginDir)
        {
            if (pluginArchive.equals(desc.getPluginArchive()))
            {
                return desc.getPluginClassname();
            }
        }
        return null;
    }

    /**
     * Get the class names of all plugins registered that are from <var>pluginDir</var> and
     * implement <var>pluginType</var>.
     * 
     * @return A set of plugin names registered from archives in <var>pluginDir</var> which
     *         implement <var>pluginType</var>.
     */
    public synchronized Set<String> getPluginClassnames(File pluginDir, Class<?> pluginType)
    {
        // create new independent copy of classnames
        Set<String> classnames = new HashSet<String>();
        Set<PluginDescriptor> pluginsInPluginDir = pluginDirToPluginSetMap.get(pluginDir);
        if (pluginsInPluginDir != null)
        {
            for (PluginDescriptor desc : pluginsInPluginDir)
            {
                if (pluginType == null || desc.isImplementing(pluginType))
                {
                    classnames.add(desc.getPluginClassname());
                }
            }
        }
        return classnames;
    }

    /**
     * Get all registered plugins.
     * 
     * @return A set of all plugin classes.
     */
    public synchronized Set<Class<?>> getPluginClasses()
    {
        final Set<Class<?>> plugins = new HashSet<Class<?>>();
        for (PluginDescriptor desc : classNameMap.values())
        {
            plugins.add(desc.getPluginClass());
        }
        return plugins;
    }

    /**
     * Get all registered plugins that implement <var>pluginType</var>.
     * 
     * @return A set of all plugin classes implementing <var>pluginType</var>.
     */
    public synchronized Set<Class<?>> getPluginClasses(Class<?> pluginType)
    {
        // create new set to prohibit deletion of plugins in original map
        final Set<Class<?>> plugins = new HashSet<Class<?>>();
        for (PluginDescriptor desc : classNameMap.values())
        {
            if (desc.isImplementing(pluginType))
            {
                plugins.add(desc.getPluginClass());
            }
        }
        return plugins;
    }

    /**
     * Get all registered plugins from <var>pluginDir</var>.
     * 
     * @return A set of all plugin classes from archives in <var>pluginDir</var>.
     */
    public Set<Class<?>> getPluginClasses(File pluginDir)
    {
        return getPluginClasses(pluginDir, null);
    }

    /**
     * Get all registered plugins from <var>pluginDir</var> that implement <var>pluginType</var>.
     * 
     * @return A set of all plugin classes from archives in <var>pluginDir</var> implementing
     *         <var>pluginType</var>.
     */
    public synchronized Set<Class<?>> getPluginClasses(File pluginDir, Class<?> pluginType)
    {
        final Set<Class<?>> plugins = new HashSet<Class<?>>();
        Set<PluginDescriptor> pluginsInPluginDir = pluginDirToPluginSetMap.get(pluginDir);
        if (pluginsInPluginDir != null)
        {
            for (PluginDescriptor desc : pluginsInPluginDir)
            {
                if (pluginType == null || desc.isImplementing(pluginType))
                {
                    plugins.add(desc.getPluginClass());
                }
            }
        }
        return plugins;
    }

    /**
     * Disable the given <var>plugin</var> by renaming the corresponding plugin archive to
     * <code>&lt;oldname&gt;.disabled</code>.
     * 
     * @param plugin The plugin class.
     * @return <code>true</code> if successfully disabled, <code>false</code> otherwise.
     */
    public synchronized boolean disablePlugin(Class<?> plugin)
    {
        if (plugin == null)
        {
            log.warn("disable plugin called with null - ignoring.");
            return false;
        }
        final PluginDescriptor desc = classNameMap.get(plugin.getName());
        if (desc == null)
        {
            log.warn("disable plugin called with class " + plugin.getName()
                    + " that is not a plugin - ignoring.");
            return false;
        }
        unregister(desc);
        return loader.disablePlugin(plugin);
    }

    /**
     * Disable the given <var>pluginArchive</var> by renaming it to
     * <code>&lt;oldname&gt;.disabled</code>.
     * 
     * @param pluginArchive The archive to get the plugin from.
     * @return <code>true</code> if successfully disabled, <code>false</code> otherwise.
     */
    public synchronized boolean disablePlugin(File pluginArchive)
    {
        if (pluginArchive == null)
        {
            log.warn("disable plugin called with null - ignoring.");
            return false;
        }
        final Set<PluginDescriptor> pluginsInPluginDir =
                pluginDirToPluginSetMap.get(pluginArchive.getParentFile());
        if (pluginsInPluginDir != null)
        {
            for (PluginDescriptor desc : pluginsInPluginDir)
            {
                if (desc.getPluginArchive().equals(pluginArchive))
                {
                    unregister(desc);
                    break;
                }
            }
        }
        return loader.disablePlugin(pluginArchive);
    }

    /**
     * Enables the plugin defined in the given <var>pluginArchive</var> by renaming it from
     * <code>&lt;oldname&gt;.jar.disabled</code> to <code>&lt;oldname&gt;.jar</code>.
     * 
     * @param pluginArchive The archive to get the plugin from.
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     * @return <code>true</code> if successfully enabled, <code>false</code> otherwise.
     */
    public boolean enablePlugin(File pluginArchive, boolean synchronous)
    {
        if (pluginArchive == null)
        {
            log.warn("enable plugin called with null - ignoring.");
            return false;
        }
        if (enableArchiveFile(pluginArchive) == false)
        {
            return false;
        }
        refresh(synchronous);
        return true;
    }

    private boolean enableArchiveFile(File pluginArchive)
    {
        // if the archive path doesn't exist or isn't a file, it is already disabled
        if (pluginArchive.isFile() && pluginArchive.getName().endsWith(".jar"))
        {
            return true;
        }
        final File disabledPluginArchive = new File(pluginArchive.getAbsolutePath() + ".disabled");
        if (disabledPluginArchive.isFile() == false)
        {
            return false;
        }
        return disabledPluginArchive.renameTo(pluginArchive);
    }

    /**
     * Shutdown plugin hot deployment for the given <var>containerName</var>.
     */
    public static void shutdown(String containerName)
    {
        final PluginContainer instance = instances.get(containerName);
        if (instance == null)
        {
            return;
        }
        instance.shutdownInstance();
        instances.remove(containerName);
    }

    /**
     * Shutdown plugin hot deployment for the main container.
     */
    public static void shutdown()
    {
        shutdown(MAIN_PLUGIN_CONTAINER_NAME);
    }

    /**
     * Shutdown plugin hot deployment for all container.
     */
    public static void shutdownAll()
    {
        for (PluginContainer instance : instances.values())
        {
            instance.shutdownInstance();
        }
        instances.clear();
    }

    private void shutdownInstance()
    {
        long begin = System.nanoTime();

        // stop plugin loader thread
        loader.notifyShutdown();
        log.info("Interrupting loader thread...");
        loaderThread.interrupt();
        synchronized (this)
        {
            try
            {
                loaderThread.join(2 * MAX_TIME_IN_SECS_TO_WAIT_FOR_LOAD_CYCLE * 1000L);
            } catch (InterruptedException ie)
            {
                log.info("Interrupted while waiting for plugin loader thread to die.");
            }
        }
        loaderThread = null;
        long middle = System.nanoTime();

        classNameMap.clear();
        pluginNameMap.clear();
        pluginDirToPluginSetMap.clear();
        // notify observers to release references
        classNameMap = new HashMap<String, PluginDescriptor>();

        notifier.notifyListeners(new PluginEvent(PluginEventType.UNREGISTER_PLUGIN, null, null));
        notifier.clear();
        notifier = null;

        // free resources
        classNameMap = null;
        pluginNameMap = null;
        pluginDirToPluginSetMap = null;
        log.info("All resources released.");

        // remove all loaded plugins
        loader.destroy();
        loader = null;

        // before we leave run gc and finalization again
        if (runGcOnShutdown)
        {
            System.gc();
            System.runFinalization();
        }

        long end = System.nanoTime();
        String msMiddle = String.valueOf((middle - begin) / 1E6);
        String msTotal = String.valueOf((end - begin) / 1E6);
        log.info("PluginContainer finished (" + msMiddle + "ms/" + msTotal + "ms).");
    }

}
