/*
 * Copyright 2005-2009 Lukas Blunschi
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.hotdeploy;

import java.io.File;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * @author Lukas Blunschi
 */
final class PluginDeploymentDesc
{
    private final Log log = LogFactory.getLog(PluginDeploymentDesc.class);

    /** the plugin archive. this may be a file or a directory. */
    private final File pluginArchive;

    /**
     * If <code>true</code>, only classes with {@link PluginInfo} annotation are considered plugins.
     */
    private final boolean requirePluginInfo;

    /** the URLClassLoader used to load this plugin */
    private ClosableURLClassLoader classLoader;

    /**
     * The immutable plugin descriptor.
     */
    private PluginDescriptor pluginDesc;

    /**
     * The constructor.<br>
     * It tries to load the plugin. A succesful loading can be checked using
     * isLoaded().
     * 
     * @param pluginArchive the plugin archive.
     */
    PluginDeploymentDesc(File pluginArchive, boolean requirePluginInfo)
    {
        this.pluginArchive = pluginArchive;
        this.requirePluginInfo = requirePluginInfo;
        this.classLoader = null;
        this.pluginDesc = null;
        reload();
    }

    /**
     * reload this plugin
     * <i>Only call in synchronized block.</i>
     * 
     * @return true if successfully reloaded.
     */
    boolean reload()
    {
        try
        {
            // try to unload
            unload();

            // ok, we can reload the plugin - let's reinit the other variables

            // create a new classloader
            // use current ClassLoader as parent
            classLoader =
                    new ClosableURLClassLoader(pluginArchive.getAbsolutePath(), getClass()
                            .getClassLoader());

            // look for plugin class
            String pluginClassname = tryLoadingFromJar(pluginArchive);

            // load plugin
            if (pluginClassname == null)
            {
                // no plugin class found - nothing to load
                unload();
            } else
            {
                final long lastModified = pluginArchive.lastModified();
                final Class<?> pluginClass = classLoader.loadClass(pluginClassname);
                pluginDesc = new PluginDescriptor(pluginClass, pluginArchive, lastModified);
                // plugin successfully loaded
            }
        } catch (Exception e)
        {
            log.error("error loading plugin: " + e.getMessage(), e);
            unload();
        }
        return pluginDesc != null;
    }

    /**
     * Unload this plugin. This is a hard unload - this object will not be used
     * anymore afterwards.
     * <i>Only call in synchronized block.</i>
     */
    void unload()
    {
        try
        {
            // first close jar file
            if (classLoader != null)
            {
                classLoader.close();
            }
            classLoader = null;

            // free the (hopefully) last reference to the plugin class
            pluginDesc = null;
        } catch (Exception e)
        {
            log.error("Failure while unloading plugin: " + e.getMessage(), e);
        }
    }

    /**
     * <i>Only call in synchronized block.</i>
     * 
     * @return true if plugin could be loaded
     */
    boolean isLoaded()
    {
        return pluginDesc != null;
    }

    /**
     * <i>Only call in synchronized block.</i>
     * 
     * @return true if plugin archive has changed.
     */
    boolean hasChanged()
    {
        return (pluginDesc == null) ? false : pluginDesc.hasPluginArchiveChanged();
    }

    /**
     * <i>Only call in synchronized block.</i>
     * 
     * @return the plugin class or null if plugin not loaded.
     */
    Class<?> getPluginClass()
    {
        return (pluginDesc == null) ? null : pluginDesc.getPluginClass();
    }

    File getPluginArchive()
    {
        return pluginArchive;
    }

    /**
     * <i>Only call in synchronized block.</i>
     * 
     * @return true if the plugin has been updated
     */
    boolean updatePluginIfChanged()
    {
        try
        {
            if (hasChanged())
            {
                if (reload())
                {
                    log.info("reload archive " + pluginArchive.getName());
                    return true;
                } else
                {
                    log.warn("reload archive " + pluginArchive.getName() + " failed.");
                    return false;
                }
            } else
            {
                return false;
            }
        } catch (RuntimeException ex)
        {
            log.error("Error: " + ex.getMessage(), ex);
            ex.printStackTrace();
            return false;
        }
    }

    /**
     * @return the plugin class or null if plugin not loaded.
     */
    synchronized Class<?> checkAndGetPluginClass()
    {
        try
        {
            if (hasChanged())
            {
                if (reload())
                {
                    log.info("reload archive " + pluginArchive.getName());
                } else
                {
                    log.warn("reload archive " + pluginArchive.getName() + " failed.");
                }
            }
            return (pluginDesc == null) ? null : pluginDesc.getPluginClass();
        } catch (RuntimeException ex)
        {
            log.error("Error: " + ex.getMessage(), ex);
            ex.printStackTrace();
            return null;
        }
    }

    /**
     * @param className the classname of the class to test
     * @return true if the given class implements the org.imemex.plugin.Plugin
     *         interface
     * @throws Exception
     */
    private boolean isPluginClass(String className) throws ClassNotFoundException
    {
        Class<?> clazz = classLoader.loadClass(className);
        // we are looking for real instantiable classes!
        // not interfaces
        if (clazz.isInterface())
        {
            return false;
        }
        return requirePluginInfo ? hasValidPluginInfo(clazz) : (Plugin.class
                .isAssignableFrom(clazz) || hasValidPluginInfo(clazz));
    }

    private boolean hasValidPluginInfo(Class<?> clazz)
    {
        PluginInfo info = clazz.getAnnotation(PluginInfo.class);
        if (info == null)
        {
            return false;
        }
        final boolean validType = info.pluginType().isAssignableFrom(clazz);
        if (validType == false)
        {
            log.warn("plugin " + clazz.getName() + " does not implement its plugin type "
                    + info.pluginType() + " - ignored");
        }
        return validType;
    }

    /**
     * try to load a plugin from the given jar file.
     * 
     * @param jar
     * @return the plugin class name or null if plugin class was not found.
     * @throws Exception
     */
    private String tryLoadingFromJar(File jar) throws Exception
    {
        String pluginClassname = null;
        ZipFile zipFile = null;
        try
        {
            zipFile = new ZipFile(jar);
            Enumeration<? extends ZipEntry> en = zipFile.entries();
            while (en.hasMoreElements())
            {
                ZipEntry entry = en.nextElement();
                String resource_name = entry.getName();
                if (resource_name.endsWith(".class"))
                {
                    // get classname (e.g. org.hotdeploy.Name)
                    // from zip entry (org/hotdeploy/Name.class)
                    String className =
                            resource_name.substring(0, resource_name.lastIndexOf(".class"));
                    className = className.replace('/', '.');
                    if (isPluginClass(className))
                    {
                        pluginClassname = className;
                        break;
                    }
                }
            }
        } finally
        {
            if (zipFile != null)
            {
                zipFile.close();
            }
        }

        return pluginClassname;
    }

    /**
     * Returns the plugin description.
     */
    PluginDescriptor getPluginDesc()
    {
        return pluginDesc;
    }

    ClosableURLClassLoader getClassLoader()
    {
        return classLoader;
    }

}
