/*
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.hotdeploy;

import java.io.File;

/**
 * A retriever for information about a given plugin. Can be used to register listeners who want to
 * get events for the given plugin.
 * 
 * @author Bernd Rinn
 */
public class PluginInfoProvider
{
    final PluginContainer container;

    final String pluginName;

    final boolean isClassName;

    private final PluginEventNotifier notifier;

    /**
     * Create an info retriever.
     * 
     * @param container The plugin container to get the plugin from.
     * @param pluginName The name of the plugin to get info (and events) for.
     */
    public PluginInfoProvider(final PluginContainer container, final String pluginName)
    {
        this(container, pluginName, false);
    }

    /**
     * Create an info provider.
     * 
     * @param container The plugin container to get the plugin from.
     * @param pluginName The name or class name of the plugin to get info (and events) for.
     * @param isClassName If <code>true</code>, <var>pluginName</var> will be considered a class
     *            name, else it is considered a plugin name.
     */
    public PluginInfoProvider(final PluginContainer container, final String pluginName,
            final boolean isClassName)
    {
        this.container = container;
        this.pluginName = pluginName;
        this.isClassName = isClassName;
        this.notifier = new PluginEventNotifier();
        container.addListener(new PluginEventListener()
            {
                @Override
                public void pluginChanged(PluginEvent event)
                {
                    if (event.isShutdownEvent()
                            || (isClassName ? pluginName.equals(event.getPluginClassname())
                                    : pluginName.equals(event.getPluginName())))
                    {
                        processEvent(event);
                    }
                }
            });
    }

    /**
     * Override this in sub-classes to be notified of events for this plugin.
     */
    protected void processEvent(PluginEvent event)
    {
        notifier.notifyListeners(event);
    }

    /**
     * Adds a listener for changes to the plugin of this plugin info provider.
     */
    public void addListener(PluginEventListener listener)
    {
        notifier.addListener(listener);
    }

    /**
     * Removes a listener for changes to the plugin of this plugin info provider.
     */
    public void removeListener(PluginEventListener listener)
    {
        notifier.removeListener(listener);
    }

    /**
     * Returns <code>true</code> if the plugin has a constructor for the
     * <var>constructorArgumentTypes</var>.
     */
    public boolean hasConstructor(Class<?>... constructorArgumentTypes)
    {
        return container.hasConstructorByClassname(pluginName, constructorArgumentTypes);
    }

    /**
     * Returns <code>true</code> if the plugin has a constructor for the <var>constructorArgs</var>.
     */
    public boolean hasConstructor(Object... constructorArgs)
    {
        return container.hasConstructorByClassname(pluginName,
                ReflectionUtils.getClasses(constructorArgs));
    }

    /**
     * Returns the class of this plugin.
     */
    public Class<?> tryGetPluginClass()
    {
        return isClassName ? container.tryGetPluginClassByClassname(pluginName) : container
                .tryGetPluginClass(pluginName);
    }

    /**
     * Returns the name of the plugin class.
     */
    public String tryGetPluginClassname()
    {
        if (isClassName)
        {
            return pluginName;
        } else
        {
            final Class<?> clazz = container.tryGetPluginClass(pluginName);
            return clazz == null ? null : clazz.getName();
        }
    }

    /**
     * Returns the name of the plugin. If the plugin has a {@link PluginInfo}, it will be
     * {@link PluginInfo#name()}, otherwise it will be the name of the plugin class.
     */
    public String getPluginName()
    {
        if (isClassName)
        {
            final PluginInfo info = container.tryGetPluginInfoByClassname(pluginName);
            if (info != null)
            {
                return info.name();
            } else
            {
                return pluginName;
            }
        } else
        {
            return pluginName;
        }
    }

    /**
     * Returns the type of the plugin. If the plugin has a {@link PluginInfo}, it will be
     * {@link PluginInfo#pluginType()}, otherwise it will be the plugin class itself.
     */
    public Class<?> getPluginType()
    {
        final PluginInfo info =
                isClassName ? container.tryGetPluginInfoByClassname(pluginName) : container
                        .tryGetPluginInfo(pluginName);
        if (info != null)
        {
            return info.pluginType();
        } else
        {
            return tryGetPluginClass();
        }
    }

    /**
     * Returns the major version of the plugin, or -1, if no major version is defined.
     */
    public int getMajorVersion()
    {
        final PluginInfo info = container.tryGetPluginInfoByClassname(pluginName);
        if (info != null)
        {
            return info.majorVersion();
        } else
        {
            return -1;
        }
    }

    /**
     * Returns the minor version of the plugin, or -1, if no minor version is defined.
     */
    public int getMinorVersion()
    {
        final PluginInfo info = container.tryGetPluginInfoByClassname(pluginName);
        if (info != null)
        {
            return info.minorVersion();
        } else
        {
            return -1;
        }
    }

    /**
     * Returns the archive file that this plugin is defined in.
     */
    public File tryGetPluginArchive()
    {
        return container.tryGetPluginArchiveByClassname(pluginName);
    }

}
