/*
 * Copyright 2005-2009 Lukas Blunschi
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.hotdeploy;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicReference;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The <code>PluginLoader</code> scans the plugin directory for new plugin archives and loads them.
 * It also removes plugins which are no longer deployed.
 * <p>
 * There should only ever be one <code>PluginLoader</code> scanning for changes, thus this class is
 * a singleton. This is ensured by the {@PluginContainer} being the only class
 * that has a reference to the loader and being a singleton itself.
 * <p>
 * The <code>PluginLoader</code> has control over the plugins. It can disable plugins by renaming
 * it. Plugins can be retrieved by archive name. This is the name of the jar file (without path
 * components)
 * <p>
 * <b>Most important:</b> The <code>PluginLoader</code> does <i>not</i> hand out any reference to a
 * plugin. Therefore it doesn't have to worry about other objects holding references to plugins. The
 * <code>PluginLoader</code> thus can't have any observers.
 * 
 * @author Lukas Blunschi
 * @author Bernd Rinn
 */
final class PluginLoader implements Runnable
{
    /**
     * The logger.
     */
    private static final Log log = LogFactory.getLog(PluginLoader.class);

    //
    // Immutable parameters
    //

    /**
     * The time interval between directory scans (in milli-seconds).
     */
    private final long scanInterval;

    /**
     * If <code>true</code>, only classes with {@link PluginInfo} annotation are considered plugins.
     */
    private final boolean requirePluginInfo;

    //
    // Internal state, only accessed by the PluginLoader thread.
    //

    /**
     * We need a reference to the plugin container to notify it of plugin registration, update and
     * un-registration.
     */
    private PluginContainer container;

    /**
     * This map contains all the currently deployed plugins. It maps from archive file to the
     * {@link PluginDeploymentDesc}.
     */
    private Map<File, PluginDeploymentDesc> pluginDescriptions;

    /**
     * A map from plugin directory to a set of archives containing a plugin found therein.
     */
    private Map<File, Set<File>> pluginDirToArchiveSetMap;

    /**
     * Maps from file path to last modified (ms) of jar files which have been found invalid.
     */
    private Map<String, Long> rejectedJarFiles;

    //
    // State that is changed by the PluginContainer from other threads than the PluginLoader thread.
    //

    /**
     * This status variable is true as long as the plugin loader shall keep running. When it gets
     * set to <code>false</code>, the <code>PluginLoader</code> thread will exit.
     * <p>
     * <i>Thread-safety strategy: atomic access.</i>
     */
    private volatile boolean keepRunning;

    // Thread-safety information:
    // All methods reading or writing pluginDirectories <i>must</i> hold the lock on the monitor
    // of this plugin container instance.

    /**
     * The directories which are scanned for plugins. Put your plugin archives in one of them.
     * <p>
     * <i>Thread-safety strategy: only access in synchronized blocks.</i>
     */
    private final Set<File> pluginDirectories;

    /**
     * This reference to a count down latch is used by {@link PluginContainer#refresh(boolean)} for
     * synchronous refreshs.
     * <p>
     * <i>Thread-safety strategy: atomic access.</i>
     */
    private final AtomicReference<CountDownLatch> syncRef;

    /**
     * This set contains plugin archives which should be disabled by renaming it to
     * <code>&lt;oldname&gt;.disabled</code>. We need this set as on Windows it may not work the
     * first time we try it as another program may hold a mandatory lock.
     * <p>
     * <i>Thread-safety strategy: synchronized set.</i>
     */
    private Set<File> toBeDisabled;

    /**
     * Creates and initializes the PluginLoader. Sets parameters scan interval and scan directory.
     * <p>
     * Supposed to be only called once from a singleton {@link PluginContainer} class.
     * 
     * @param container
     * @param params
     */
    PluginLoader(PluginContainer container, PluginContainerParameters params)
    {
        this.scanInterval = params.getScanIntervalMillis();
        this.requirePluginInfo = params.isRequirePluginInfo();
        log.info("plugin scan interval set to: " + scanInterval + "ms.");
        this.container = container;
        this.pluginDescriptions = new HashMap<File, PluginDeploymentDesc>();
        this.pluginDirToArchiveSetMap = new HashMap<File, Set<File>>();
        this.rejectedJarFiles = new HashMap<String, Long>();
        this.toBeDisabled = Collections.synchronizedSet(new HashSet<File>());
        this.keepRunning = true;
        this.syncRef = new AtomicReference<CountDownLatch>();
        this.pluginDirectories = new LinkedHashSet<File>();
        this.pluginDirectories.addAll(params.getPluginDirs());
        log.info("plugin directories set to: " + pluginDirectories);
    }

    /**
     * Returns a set with all currently registered plugin directories.
     * 
     * @return A copy of the set of plugin directories.
     */
    Set<File> getPluginDirectories()
    {
        final Set<File> copy = new HashSet<File>();
        synchronized (pluginDirectories)
        {
            copy.addAll(pluginDirectories);
        }
        return copy;
    }

    /**
     * Add <var>pluginDirectory</var> to the search lists of plugins. Does not wait for the refresh
     * cycle to be finished.
     */
    void addPluginDirectory(File pluginDirectory)
    {
        synchronized (pluginDirectories)
        {
            pluginDirectories.add(pluginDirectory);
        }
    }

    /**
     * Add all <var>pluginDirectoriesToBeAdded</var> to the search lists of plugins. Does not wait
     * for the refresh cycle to be finished.
     */
    void addPluginDirectories(Collection<File> pluginDirectoriesToBeAdded)
    {
        synchronized (pluginDirectories)
        {
            pluginDirectories.addAll(pluginDirectoriesToBeAdded);
        }
    }

    /**
     * Remove <var>pluginDirectory</var> from the search lists of plugins. Does not wait for the
     * refresh cycle to be finished.
     */
    void removePluginDirectory(File pluginDirectory)
    {
        synchronized (pluginDirectories)
        {
            pluginDirectories.remove(pluginDirectory);
        }
    }

    /**
     * Remove all <var>pluginDirectoriesToBeRemoved</var> from the search lists of plugins. Does not
     * wait for the refresh cycle to be finished.
     */
    void removePluginDirectories(Collection<File> pluginDirectoriesToBeRemoved)
    {
        synchronized (pluginDirectories)
        {
            pluginDirectories.removeAll(pluginDirectoriesToBeRemoved);
        }
    }

    /**
     * Sets the <var>pluginDirectories</var> as the search lists of plugins, removing all plugin
     * directories set previously. Does not wait for the refresh cycle to be finished.
     */
    void setPluginDirectories(Collection<File> newPluginDirectories)
    {
        synchronized (pluginDirectories)
        {
            pluginDirectories.clear();
            pluginDirectories.addAll(newPluginDirectories);
        }
    }

    /**
     * The run function of this {@link Runnable}.
     * <p>
     * In the first loop, will loop over all jar files in all plugin directories, looking for new
     * and changed plugins. In the second loop it will remove any plugin where it's archive doesn't
     * exist anymore.
     */
    @Override
    public void run()
    {
        log.info("PluginLoader thread started.");

        final Set<File> archivesSeen = new HashSet<File>();
        CountDownLatch sync = syncRef.getAndSet(null);

        while (keepRunning)
        {
            try
            {
                long begin = System.nanoTime();

                // clear the set of archives we have seen in this run
                archivesSeen.clear();

                // cleanup and get scanned plugin directories
                final List<File> currentPluginDirs = getPluginDirsWithCleanup();
                if (log.isDebugEnabled())
                {
                    log.debug("scanned plugin directories: " + currentPluginDirs);
                }
                for (File pluginDir : currentPluginDirs)
                {
                    // loop over all archives in the plugin dir
                    File[] pluginArchives = pluginDir.listFiles();
                    if (log.isDebugEnabled())
                    {
                        log.debug("entries found in directory " + pluginDir.getCanonicalPath()
                                + ": " + Arrays.toString(pluginArchives));
                    }
                    if (pluginArchives == null)
                    {
                        continue;
                    }
                    for (File pluginArchive : pluginArchives)
                    {
                        if (checkArchiveIsRejected(pluginArchive)
                                || checkArchiveIsDisabled(pluginArchive)
                                || checkPathIsNoJarFile(pluginArchive))
                        {
                            continue;
                        }
                        archivesSeen.add(pluginArchive);
                        if (checkPathIsInvalidJarFile(pluginArchive))
                        {
                            continue;
                        }
                        processPlugin(pluginArchive);
                    } // for pluginArchive
                } // for pluginDir

                unloadAndUnregisterVanishedArchives(archivesSeen);

                if (log.isDebugEnabled())
                {
                    final String ms = String.valueOf((System.nanoTime() - begin) / 1E6);
                    log.debug("plugin loader finished run in " + ms + "ms");
                }

            } catch (Throwable e)
            {
                if (e.getMessage() != null)
                {
                    log.error("Unhandled exception thrown in PluginLoader: " + e.getMessage(), e);
                } else
                {
                    log.error("Unhandled exception thrown in PluginLoader: ", e);
                }
            } finally
            {
                if (sync != null)
                {
                    sync.countDown();
                }
                if (keepRunning == false)
                {
                    sync = syncRef.get();
                    if (sync != null)
                    {
                        sync.countDown();
                    }
                    break;
                }
                // check for loader thread interrupted while processing
                if (Thread.interrupted() == false)
                {
                    sleep();
                }
                sync = syncRef.getAndSet(null);
            }
        }
        log.info("PluginLoader thread finished.");
    }

    interface ClassLoaderCallable<V> {
        V call(ClassLoader classLoader) throws Exception;
    }

    /**
     * Runs the given <var>callable</var> in a synchronized block of the the deployment descriptor
     * of <var>pluginArchive</var>. Returns <code>null</code> if either the plugin archive's
     * descriptor cannot be found or the call fails.
     */
    public <T> T tryRunLocked(File pluginArchive, ClassLoaderCallable<T> callable)
    {
        final PluginDeploymentDesc desc = pluginDescriptions.get(pluginArchive);
        if (desc == null)
        {
            return null;
        }
        synchronized (desc)
        {
            if (desc.isLoaded() == false)
            {
                return null;
            }
            try
            {
                return callable.call(desc.getClassLoader());
            } catch (Exception ex)
            {
                log.error("Error running callable");
                return null;
            }
        }
    }

    /**
     * Sleep for <var>scanInterval</var> ms, or until this thread is interrupted.
     */
    private synchronized void sleep()
    {
        try
        {
            if (scanInterval < 0)
            {
                // loop to cover up for 'spurious wake-ups'
                while (true)
                {
                    this.wait();
                }
            } else
            {
                Thread.sleep(scanInterval);
            }
        } catch (InterruptedException ie)
        {
        }
    }

    /**
     * Returns a copy of the plugin directories, removing all plugins from plugin directories which
     * have been removed.
     */
    private List<File> getPluginDirsWithCleanup()
    {
        final List<File> currentPluginDirs = new ArrayList<File>(pluginDirectories.size());
        synchronized (pluginDirectories)
        {
            currentPluginDirs.addAll(pluginDirectories);
            final List<File> toBeRemoved = new ArrayList<File>();
            // unregister all plugins in plugin dirs which have been removed;
            // do this before we process new plugins to avoid a new plugin to be falsely rejected
            for (Entry<File, Set<File>> p : pluginDirToArchiveSetMap.entrySet())
            {
                final File pluginDir = p.getKey();
                final Set<File> archives = p.getValue();
                if (pluginDirectories.contains(pluginDir) == false && archives != null
                        && archives.isEmpty() == false)
                {
                    log.info("unregister all plugins in plugin directory " + pluginDir);
                    for (File archive : archives)
                    {
                        unloadAndUnregisterArchive(archive);
                    }
                    toBeRemoved.add(pluginDir);
                }
            }
            // some rejected jar files may be accepted in the absence of the archives to be removed
            if (toBeRemoved.isEmpty() == false)
            {
                rejectedJarFiles.clear();
            }
            // remove them from the pluginDirToArchiveSetMap after iterating over the entries
            // to avoid a ConcurrentModificationException
            for (File pluginDir : toBeRemoved)
            {
                pluginDirToArchiveSetMap.remove(pluginDir);
            }
        }
        return currentPluginDirs;
    }

    /**
     * Unload and unregister all plugins from archives that have vanished since the last scan.
     */
    private void unloadAndUnregisterVanishedArchives(final Set<File> archivesSeen)
    {
        final Collection<File> tmpToRemove = new ArrayList<File>();
        for (File archive : pluginDescriptions.keySet())
        {
            if (archivesSeen.contains(archive) == false)
            {
                tmpToRemove.add(archive);
            }
        }
        for (File archive : tmpToRemove)
        {
            unloadAndUnregisterArchive(archive);
        }
    }

    /**
     * Unloads and unregisters <var>archiveName</var>.
     */
    private void unloadAndUnregisterArchive(File archive)
    {
        final PluginDeploymentDesc desc = pluginDescriptions.remove(archive);
        if (desc == null)
        {
            return;
        }
        synchronized (desc)
        {
            if (desc.isLoaded())
            {
                try
                {
                    container.unregister(desc.getPluginDesc());
                    desc.unload();
                    log.info("remove archive " + archive);
                } catch (Exception pe)
                {
                    log.error("removal of plugin " + archive + " failed: " + pe.getMessage(), pe);
                }
            }
        }
    }

    /**
     * Process the plugin <var>pluginArchive</var>, finding out whether it is new or not.
     */
    private PluginDeploymentDesc processPlugin(File pluginArchive)
    {
        final PluginDeploymentDesc desc = pluginDescriptions.get(pluginArchive);
        if (desc != null)
        {
            if (pluginArchive.exists() == false)
            {
                unloadAndUnregisterArchive(pluginArchive);
            }
            processKnownPlugin(pluginArchive, desc);
        } else
        {
            processNewPlugin(pluginArchive);
        }
        return desc;
    }

    /**
     * Process the new plugin <var>pluginArchive</var>.
     */
    private void processNewPlugin(File pluginArchive)
    {
        PluginDeploymentDesc desc = new PluginDeploymentDesc(pluginArchive, requirePluginInfo);
        if (pluginConflicts(pluginArchive, desc))
        {
            rejectedJarFiles.put(pluginArchive.getPath(), pluginArchive.lastModified());
            return;
        }
        log.info("load new archive " + pluginArchive);
        if (desc.isLoaded())
        {
            try
            {
                container.register(desc.getPluginDesc(), true);
                pluginDescriptions.put(pluginArchive, desc);
                addArchiveToPluginDirMap(pluginArchive.getParentFile(), pluginArchive);
            } catch (Exception pe)
            {
                log.error("error registering plugin: " + pe.getMessage(), pe);
            }
        } else
        {
            log.error("no valid plugin found in: " + pluginArchive);
            rejectedJarFiles.put(pluginArchive.getPath(), pluginArchive.lastModified());
        }
    }

    /**
     * Process the already known plugin <var>pluginArchive</var>.
     */
    @SuppressWarnings("null")
    private void processKnownPlugin(File pluginArchive, PluginDeploymentDesc desc)
    {
        synchronized (desc)
        {
            final PluginDescriptor oldPluginDesc = desc.getPluginDesc();
            if (desc.updatePluginIfChanged())
            {
                // has the new plugin archive a plugin with a different
                // name?
                final PluginDescriptor newPluginDesc = desc.getPluginDesc();
                final boolean newPlugin =
                        (oldPluginDesc != null)
                                && oldPluginDesc.getPluginClassname().equals(
                                        newPluginDesc.getPluginClassname()) == false;
                if (newPlugin)
                {
                    container.unregister(oldPluginDesc);
                    if (log.isDebugEnabled())
                    {
                        log.debug("archive " + pluginArchive.getName() + ": replacing plugin "
                                + oldPluginDesc.getPluginClassname() + " with "
                                + newPluginDesc.getPluginClassname());
                    }
                }
                container.register(newPluginDesc, newPlugin);
            }
        }
    }

    private boolean pluginConflicts(File pluginArchive, PluginDeploymentDesc desc)
    {
        final Class<?> pluginClass = desc.getPluginClass();
        if (pluginClass == null)
        {
            return false;
        }
        PluginDescriptor registeredPluginDesc =
                container.tryGetPluginDesc(pluginClass.getName(), true, false);
        if (registeredPluginDesc != null)
        {
            log.warn("ignoring new archive " + pluginArchive + " as plugin class "
                    + desc.getPluginClass().getName() + " registered from "
                    + registeredPluginDesc.getPluginArchive());
            return true;
        }
        registeredPluginDesc =
                container.tryGetPluginDesc(desc.getPluginDesc().getPluginName(), false, false);
        if (registeredPluginDesc != null)
        {

            log.warn("ignoring new archive " + pluginArchive + " as plugin "
                    + desc.getPluginDesc().getPluginName() + " (classname: "
                    + desc.getPluginClass().getName() + ") is already registered from "
                    + registeredPluginDesc.getPluginArchive());
            return true;
        }
        return false;
    }

    /**
     * Checks if <var>pluginArchive</var> is invalid when parsed as a jar file.
     */
    private boolean checkPathIsInvalidJarFile(File pluginArchive)
    {
        if (isValidJarArchive(pluginArchive) == false)
        {
            log.warn("archive is not a valid jar file: " + pluginArchive);
            rejectedJarFiles.put(pluginArchive.getPath(), pluginArchive.lastModified());
            return true;
        }
        return false;
    }

    /**
     * Check if a file is a valid jar file.
     * 
     * @param file The file to be tested.
     * @return <code>true</code> if the file is a valid jar file, <code>false</code> otherwise.
     */
    private boolean isValidJarArchive(File file)
    {
        boolean isArchive = true;
        ZipFile zipFile = null;

        try
        {
            zipFile = new ZipFile(file);
        } catch (ZipException zipCurrupted)
        {
            isArchive = false;
        } catch (IOException anyIOError)
        {
            isArchive = false;
        } finally
        {
            if (zipFile != null)
            {
                try
                {
                    zipFile.close();
                } catch (IOException ioe)
                {
                    log.error("Error while closing zip file: " + ioe.getMessage(), ioe);
                }
            }
        }

        return isArchive;
    }

    private boolean checkPathIsNoJarFile(File pluginArchive)
    {
        if (pluginArchive.isHidden() || (pluginArchive.isFile() == false))
        {
            return true;
        }
        if (pluginArchive.getName().endsWith(".jar") == false)
        {
            return true;
        }
        return false;
    }

    private boolean checkArchiveIsRejected(File pluginArchive)
    {
        final Long rejectedJarFileLastModified = rejectedJarFiles.get(pluginArchive.getPath());
        return (rejectedJarFileLastModified != null && pluginArchive.lastModified() == rejectedJarFileLastModified);
    }

    private boolean checkArchiveIsDisabled(File pluginArchive)
    {
        if (toBeDisabled.contains(pluginArchive))
        {
            if (pluginArchive.exists())
            {
                // try to make sure all finalizations are done
                // running gc 2 times...
                System.runFinalization();
                System.gc();
                System.runFinalization();
                System.gc();

                // try to disable
                if (disableArchiveFile(pluginArchive))
                {
                    // plugin archive disabled
                    toBeDisabled.remove(pluginArchive);
                } else
                {
                    log.warn("plugin archive " + pluginArchive + " could not be disabled.");
                }
            } else
            {
                // remove obsolete entries
                toBeDisabled.remove(pluginArchive);
            }
            return true;
        }
        return false;
    }

    private void addArchiveToPluginDirMap(File pluginDir, File archive)
    {
        Set<File> archivesInDir = pluginDirToArchiveSetMap.get(pluginDir);
        if (archivesInDir == null)
        {
            archivesInDir = new HashSet<File>();
            pluginDirToArchiveSetMap.put(pluginDir, archivesInDir);
        }
        archivesInDir.add(archive);
    }

    /**
     * Delete a given plugin. This function takes the archive name as it appears in the plugin
     * directory as an argument. If you want to delete a plugin using a reference to the plugin,
     * 
     * @param archive archive file of the plugin to delete
     * @return <code>true</code> if deleted successful, <code>false</code> otherwise (e.g. if
     *         argument is <code>null</code> or plugin not found)
     */
    public boolean disablePlugin(File archive)
    {
        boolean disabled = false;
        if (archive == null)
        {
            log.warn("disable plugin called with null - ignoring.");
        } else
        {
            PluginDeploymentDesc desc = pluginDescriptions.get(archive.getName());
            if (desc == null)
            {
                log.warn("plugin with given archive name (" + archive + ") not found.");
            } else
            {
                synchronized (desc)
                {
                    // first unregister it in the container
                    container.unregister(desc.getPluginDesc());

                    // remove from map
                    pluginDescriptions.remove(archive.getName());
                    log.info("plugin from plugin archive map removed.");

                    // unload
                    desc.unload();
                }
            }

            // try to delete archivename in any case if possible
            desc = null;
            System.gc();
            System.runFinalization();
            System.gc();
            System.runFinalization();
            if (disableArchiveFile(archive))
            {
                // disabling by rename successful, make sure toBeDisabled doesn't contain this
                // archive name
                toBeDisabled.remove(archive);
            } else
            {
                // not renamed successfully, make sure toBeDisabled contains this archive name
                toBeDisabled.add(archive);
            }

            disabled = true;
        }
        return disabled;
    }

    private boolean disableArchiveFile(File archive)
    {
        // if the archive path doesn't exist or isn't a file, it is already disabled
        if (archive.isFile() == false)
        {
            return true;
        }
        return archive.renameTo(new File(archive.getAbsolutePath() + ".disabled"));
    }

    /**
     * Disable the given <var>plugin</var> by renaming it to <code>&lt;oldname&gt;.disabled</code>.
     * 
     * @param plugin
     * @return <code>true</code> if disabled successfully, <code>false</code> otherwise (e.g. if
     *         argument is <code>null</code> or plugin not found)
     */
    public boolean disablePlugin(Class<?> plugin)
    {
        if (plugin == null)
        {
            log.warn("disable plugin called with null - ignoring.");
            return false;
        }
        File archive = getArchive(plugin);
        if (archive == null)
        {
            String msg = "disable plugin: plugin not found (" + plugin.getName() + ").";
            log.error(msg, new Exception(msg));
            return false;
        } else
        {
            return disablePlugin(archive);
        }
    }

    /**
     * For a given plugin, find its archive.
     * 
     * @param pluginClass
     * @return archive name or null if not found (or argument is null)
     */
    private File getArchive(Class<?> pluginClass)
    {
        for (PluginDeploymentDesc desc : pluginDescriptions.values())
        {
            synchronized (desc)
            {
                if (desc.getPluginClass().getName().equals(pluginClass.getName()))
                {
                    return desc.getPluginArchive();
                }
            }
        }
        return null;
    }

    /**
     * Returns a count down latch that can be used to {@link CountDownLatch#await()} for the next
     * refresh cycle to be finished.
     */
    CountDownLatch getCycleFinishedBarrier()
    {
        if (keepRunning == false)
        {
            return new CountDownLatch(0); // game over - return a dummy
        }
        while (true)
        {
            final CountDownLatch sync = syncRef.get();
            if (sync != null)
            {
                return sync;
            }
            final CountDownLatch newSync = new CountDownLatch(1);
            if (syncRef.compareAndSet(sync, newSync))
            {
                return newSync;
            }
        }
    }

    /**
     * Notify the loader that it should stop, caller should interrupt the loader thread afterwards.
     */
    void notifyShutdown()
    {
        keepRunning = false;
    }

    /**
     * Unload all plugins and free up resources. It will release all resources!
     */
    void destroy()
    {
        // unload plugins
        for (PluginDeploymentDesc desc : pluginDescriptions.values())
        {
            desc.unload();
        }
        log.info("All plugin archives unloaded.");

        // release members
        pluginDescriptions.clear();
        pluginDescriptions = null;
        pluginDirToArchiveSetMap.clear();
        pluginDirToArchiveSetMap = null;
        toBeDisabled.clear();
        toBeDisabled = null;
        container = null;
        log.info("All resources released. PluginLoader finished.");
    }

}
