/*
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.hotdeploy;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * A map of all plugins that implement a given worker interface.
 * 
 * @author Bernd Rinn
 */
public class PluginMapHolder<T>
{
    private static final Log log = LogFactory.getLog(PluginMapHolder.class);

    private PluginContainer container;

    final Class<T> workerInterface;

    private final Map<String, T> instanceMap;

    private final boolean isClassName;

    private final Class<?>[][] constructorTypeAlternatives;

    private final Object[][] constructorArgsAlternatives;

    private final PluginEventNotifier notifier;

    /**
     * Create an plugin map holder, using plugin names as keys.
     * 
     * @param container The container to get the plugins from.
     * @param workerInterface The interface to cast the plugins to (only plugins implementing this
     *            interface will be held in the map).
     * @param constructorArgs The arguments to provide to the constructor.
     */
    public PluginMapHolder(final PluginContainer container, final Class<T> workerInterface,
            final Object... constructorArgs)
    {
        this(container, workerInterface, false, new Object[][]
            { constructorArgs });
    }

    /**
     * Create an plugin map holder.
     * 
     * @param container The container to get the plugins from.
     * @param workerInterface The interface to cast the plugins to (only plugins implementing this
     *            interface will be held in the map).
     * @param useClassNames If <code>true</code>, plugin class names will be used as keys for the
     *            map, otherwise plugin names will be used.
     * @param constructorArgs The arguments to provide to the constructor.
     */
    public PluginMapHolder(final PluginContainer container, final Class<T> workerInterface,
            final boolean useClassNames, final Object... constructorArgs)
    {
        this(container, workerInterface, useClassNames, new Object[][]
            { constructorArgs });
    }

    /**
     * Create an plugin map holder.
     * 
     * @param container The container to get the plugins from.
     * @param workerInterface The interface to cast the plugins to (only plugins implementing this
     *            interface will be held in the map). map, otherwise plugin names will be used.
     * @param constructorArgsAlternatives The alternatives for the arguments to provide to the
     *            constructor. The first alternative that matches will be used.
     */
    public PluginMapHolder(final PluginContainer container, final Class<T> workerInterface,
            final Object[][] constructorArgsAlternatives)
    {
        this(container, workerInterface, false, constructorArgsAlternatives);
    }

    /**
     * Create an plugin map holder.
     * 
     * @param container The container to get the plugins from.
     * @param workerInterface The interface to cast the plugins to (only plugins implementing this
     *            interface will be held in the map).
     * @param useClassnames If <code>true</code>, plugin class names will be used as keys for the
     *            map, otherwise plugin names will be used.
     * @param constructorArgsAlternatives The alternatives for the arguments to provide to the
     *            constructor. The first alternative that matches will be used.
     */
    public PluginMapHolder(final PluginContainer container, final Class<T> workerInterface,
            final boolean useClassnames, final Object[][] constructorArgsAlternatives)
    {
        this.container = container;
        this.workerInterface = workerInterface;
        this.instanceMap = Collections.synchronizedMap(new HashMap<String, T>());
        this.isClassName = useClassnames;
        this.constructorArgsAlternatives = constructorArgsAlternatives;
        this.constructorTypeAlternatives = toTypes(constructorArgsAlternatives);
        this.notifier = new PluginEventNotifier();
        container.addListener(new PluginEventListener()
            {
                @Override
                public void pluginChanged(PluginEvent event)
                {
                    if (event.isShutdownEvent())
                    {
                        if (log.isDebugEnabled())
                        {
                            log.debug("Clear plugin map");
                        }
                        instanceMap.clear();
                        return;
                    }
                    if (event.getPluginClass() != null
                            && workerInterface.isAssignableFrom(event.getPluginClass()))
                    {
                        final String key =
                                useClassnames ? event.getPluginClassname() : event.getPluginName();
                        if (event.isRegistrationOrUpdateEvent())
                        {
                            final T instance = create(event.getPluginClassname());
                            if (log.isDebugEnabled())
                            {
                                log.debug("Set new plugin instance for plugin  " + key);
                            }
                            instanceMap.put(key, instance);
                        } else if (event.isUnregistrationEvent())
                        {
                            if (log.isDebugEnabled())
                            {
                                log.debug("Remove plugin instance for plugin " + key);
                            }
                            instanceMap.remove(key);
                        }
                        notifier.notifyListeners(event);
                    }
                }
            });
        // Pre-populate map with currently registered plugin of that type.
        for (PluginDescriptor p : container.getPlugins())
        {
            if (p.isImplementing(workerInterface))
            {
                final T instance = create(p.getPluginClassname());
                final String key =
                        useClassnames ? p.getPluginClassname() : p.getPluginName();
                if (log.isDebugEnabled())
                {
                    log.debug("Set new plugin instance for plugin  " + key);
                }
                instanceMap.put(key, instance);
            }
        }
    }

    private static Class<?>[][] toTypes(Object[][] constructorArgsAlternatives)
    {
        final Class<?>[][] types = new Class<?>[constructorArgsAlternatives.length][];
        int idx = 0;
        for (Object[] constructorArgs : constructorArgsAlternatives)
        {
            types[idx++] = ReflectionUtils.getClasses(constructorArgs);
        }
        return types;
    }

    private T create(String pluginClassName)
    {
        for (int i = 0; i < constructorTypeAlternatives.length; ++i)
        {
            if (container
                    .hasConstructorByClassname(pluginClassName, constructorTypeAlternatives[i]))
            {
                return container.tryCreatePluginByClassname(pluginClassName, workerInterface,
                        constructorArgsAlternatives[i]);
            }
        }
        throw new IllegalArgumentException(
                "No suitable constructor found for plugin pluginClassName");
    }

    /**
     * Adds a listener for changes to any of the plugins that this map holder holds.
     */
    public void addListener(PluginEventListener listener)
    {
        notifier.addListener(listener);
    }

    /**
     * Removes a listener for changes to any of the plugins that this map holder holds.
     */
    public void removeListener(PluginEventListener listener)
    {
        notifier.removeListener(listener);
    }

    /**
     * Returns the plugin with given <var>name</var>.
     * 
     * @param name Either the plugin class name (if <var>useClassNames</var> is <code>true</code>,
     *            or otherwise the plugin name.
     * @return The plugin with the given <var>name</var>, or <code>null</code>, if this plugin
     *         cannot be found.
     */
    public T tryGet(String name)
    {
        return instanceMap.get(name);
    }

    /**
     * Returns the names of the available plugins.
     */
    public Set<String> getPluginNames()
    {
        return instanceMap.keySet();
    }

    /**
     * Returns the given <var>resource</var> in the jar archive of the plugin with given
     * <var>pluginName</var>, or <code>null</code> , if it cannot be found.
     */
    public byte[] tryGetResource(String pluginName, String resource)
    {
        final File pluginArchive =
                isClassName ? container.tryGetPluginArchiveByClassname(pluginName) : container
                        .tryGetPluginArchive(pluginName);
        return container.tryGetPluginResource(pluginArchive, resource);
    }

    /**
     * Refresh plugins now.
     * 
     * @param synchronous if <code>true</code>, wait for the refresh cycle to finish before
     *            returning.
     */
    public void refresh(boolean synchronous)
    {
        container.refresh(synchronous);
    }
}
