/*
 * Copyright 2012 ETH Zuerich, CISD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.ethz.cisd.hotdeploy;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import javax.tools.Diagnostic;
import javax.tools.JavaFileObject;

import org.apache.commons.io.FileUtils;

import ch.ethz.cisd.compile.SimpleJavaCompiler;
import ch.systemsx.cisd.base.exceptions.CheckedExceptionTunnel;
import ch.systemsx.cisd.base.exceptions.IOExceptionUnchecked;

/**
 * A test class for easily compiling and testing Java scripts.
 * 
 * @author Bernd Rinn
 */
public class ScriptPluginTester implements Closeable
{
    private final SimpleJavaCompiler compiler;

    private final boolean requirePluginInfo;

    private PluginTester tester;

    public ScriptPluginTester()
    {
        this(null, false);
    }

    public ScriptPluginTester(File tempRootDirectory)
    {
        this(tempRootDirectory, false);
    }

    public ScriptPluginTester(boolean requirePluginInfo)
    {
        this(null, requirePluginInfo);
    }

    public ScriptPluginTester(File tempRootDirectory, boolean requirePluginInfo)
    {
        this.compiler = new SimpleJavaCompiler(tempRootDirectory);
        this.requirePluginInfo = requirePluginInfo;
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#getLastClassName()
     */
    public String getLastClassName()
    {
        return compiler.getLastClassName();
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#getCompilerOptions()
     */
    public List<String> getCompilerOptions()
    {
        return compiler.getCompilerOptions();
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#getDiagnostics()
     */
    public List<Diagnostic<? extends JavaFileObject>> getDiagnostics()
    {
        return compiler.getDiagnostics();
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#getErrorMessage()
     */
    public String getErrorMessage()
    {
        return compiler.getErrorMessage();
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#tryGetFileObjectFromSourceCode(java.lang.String)
     */
    public JavaFileObject tryGetFileObjectFromSourceCode(String classSource)
    {
        return compiler.tryGetFileObjectFromSourceCode(classSource);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#tryGetFileObjectFromSourceCode(java.lang.String,
     *      java.lang.String)
     */
    public JavaFileObject tryGetFileObjectFromSourceCode(String classNameOrNull, String classSource)
    {
        return compiler.tryGetFileObjectFromSourceCode(classNameOrNull, classSource);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#tryGetFileObjectForFile(java.io.File)
     */
    public JavaFileObject tryGetFileObjectForFile(File javaFile)
    {
        return compiler.tryGetFileObjectForFile(javaFile);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#getFileObjectsForDirectory(java.io.File)
     */
    public Iterable<? extends JavaFileObject> getFileObjectsForDirectory(File directory)
    {
        return compiler.getFileObjectsForDirectory(directory);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#compile(java.lang.Iterable)
     */
    public boolean compile(Iterable<? extends JavaFileObject> fileObjects)
            throws IOExceptionUnchecked
    {
        checkTesterNull();
        return compiler.compile(fileObjects);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#compile(javax.tools.JavaFileObject[])
     */
    public boolean compile(JavaFileObject... fileObjects) throws IOExceptionUnchecked
    {
        checkTesterNull();
        return compiler.compile(fileObjects);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#addFile(java.io.File, java.lang.String)
     */
    public void addFile(File source, String destPath) throws IOExceptionUnchecked
    {
        checkTesterNull();
        compiler.addFile(source, destPath);
    }

    /**
     * @see ch.ethz.cisd.compile.SimpleJavaCompiler#addContent(java.io.InputStream,
     *      java.lang.String)
     */
    public void addContent(InputStream istream, String destPath) throws IOExceptionUnchecked
    {
        checkTesterNull();
        compiler.addContent(istream, destPath);
    }

    /**
     * Loads the plugin.
     * <p>
     * <i>Call this method after all classes needed for the plugin have been compiled.</i>
     * 
     * @return The {@link PluginTester}, or <code>null</code>, if either the jar file could not be
     *         created or the jar file did not contain a valid plugin. Call
     *         {@link #getErrorMessage()} to see what went wrong.
     */
    public PluginTester tryLoadPlugin()
    {
        checkTesterNull();
        final File pluginArchive = compiler.tryCreateJar(compiler.getTempDir());
        return tryCreatePluginTester(pluginArchive);
    }

    /**
     * Convenience method for compiling a plugin from all Java source files in
     * <var>sourceDirectory</var>, creating a jar file and loading it as a plugin.
     * 
     * @param sourceDirectory The directory where the java files are found.
     * @return The PluginTester, or <code>null</code>, if either compilation, jar creation or plugin
     *         loading failed. Call {@link #getErrorMessage()} to see what went wrong.
     */
    public PluginTester tryCompileAndLoadFromFiles(File sourceDirectory)
            throws IOExceptionUnchecked
    {
        checkTesterNull();
        final Iterable<? extends JavaFileObject> fileObjects =
                getFileObjectsForDirectory(sourceDirectory);
        if (compile(fileObjects) == false)
        {
            return null;
        }
        final File pluginArchive = compiler.tryCreateJar(compiler.getTempDir());
        return tryCreatePluginTester(pluginArchive);
    }

    /**
     * Convenience method for compiling a plugin from all Java source files in
     * <var>sourceDirectory</var>, creating a jar file, loading it as a plugin and creating an
     * instance of the plugin.
     * 
     * @param sourceDirectory The directory where the java files are found.
     * @param workerInterface The Java interfac to cast the plugin to.
     * @param constructorArguments The arguments to provide to the constructor.
     * @return The plugin instance, or <code>null</code>, if either compilation, jar creation or
     *         plugin loading failed. Call {@link #getErrorMessage()} to see what went wrong.
     */
    public <T> T tryCompileAndCreateFromFiles(File sourceDirectory, Class<T> workerInterface,
            Object... constructorArguments) throws IOExceptionUnchecked
    {
        checkTesterNull();
        final Iterable<? extends JavaFileObject> fileObjects =
                getFileObjectsForDirectory(sourceDirectory);
        if (compile(fileObjects) == false)
        {
            return null;
        }
        final File pluginArchive = compiler.tryCreateJar(compiler.getTempDir());
        if (tryCreatePluginTester(pluginArchive) == null)
        {
            return null;
        }
        return tester.tryCreatePlugin(workerInterface, constructorArguments);
    }

    /**
     * Convenience method for compiling a plugin from a given <var>classSource</var>, creating a jar
     * file and loading it as a plugin.
     * 
     * @param classSource The directory where the java files are found.
     * @return The PluginTester, or <code>null</code>, if either compilation, jar creation or plugin
     *         loading failed. Call {@link #getErrorMessage()} to see what went wrong.
     */
    public PluginTester tryCompileAndLoadFromString(String classSource) throws IOExceptionUnchecked
    {
        checkTesterNull();
        final JavaFileObject fileObject = tryGetFileObjectFromSourceCode(classSource);
        if (fileObject == null)
        {
            return null;
        }
        if (compile(fileObject) == false)
        {
            return null;
        }
        final File pluginArchive = compiler.tryCreateJar(compiler.getTempDir());
        return tryCreatePluginTester(pluginArchive);
    }

    /**
     * Convenience method for compiling a plugin from a given <var>classSource</var>, creating a jar
     * file, loading it as a plugin and creating an instance of the plugin.
     * 
     * @param classSource The directory where the java files are found.
     * @param workerInterface The Java interfac to cast the plugin to.
     * @param constructorArguments The arguments to provide to the constructor.
     * @return The plugin instance, or <code>null</code>, if either compilation, jar creation or
     *         plugin loading failed. Call {@link #getErrorMessage()} to see what went wrong.
     */
    public <T> T tryCompileAndCreateFromString(String classSource, Class<T> workerInterface,
            Object... constructorArguments) throws IOExceptionUnchecked
    {
        checkTesterNull();
        final JavaFileObject fileObject = tryGetFileObjectFromSourceCode(classSource);
        if (fileObject == null)
        {
            return null;
        }
        if (compile(fileObject) == false)
        {
            return null;
        }
        final File pluginArchive = compiler.tryCreateJar(compiler.getTempDir());
        if (tryCreatePluginTester(pluginArchive) == null)
        {
            return null;
        }
        return tester.tryCreatePlugin(workerInterface, constructorArguments);
    }

    /**
     * Get the plugin tester. It needs to have been created by {@link #tryLoadPlugin()} or any of
     * the convenience methods which do compilation and plugin loading like
     * {@link #tryCompileAndCreateFromFiles(File, Class, Object...)}.
     * 
     * @return The plugin tester for further testing the plugin.
     * @throw IllegalStateException If trying to call this method before the plugin has been
     *        created.
     */
    public PluginTester getPluginTester()
    {
        if (tester == null)
        {
            throw new IllegalStateException(
                    "Cannot return the plugin tester before it has been been created.");
        }
        return tester;
    }

    /**
     * Makes a permanent copy of the jar file. Note that the jar file will be automatically deleted
     * when {@link #close()} is called.
     * 
     * @param destinationDir The directory to copy the jar file to.
     * @return The copied jar file.
     * @throws IllegalStateException If the jar file has not yet been created.
     * @throws IOExceptionUnchecked If there was an error with copying the jar file.
     */
    public File copyPluginJar(File destinationDir) throws IOExceptionUnchecked
    {
        return copyPluginJar(destinationDir, null);
    }

    /**
     * Makes a permanent copy of the jar file. Note that the jar file will be automatically deleted
     * when {@link #close()} is called.
     * 
     * @param destinationDir The directory to copy the jar file to.
     * @param name The new name of the jar file.
     * @return The copied jar file.
     * @throws IllegalStateException If the jar file has not yet been created.
     * @throws IOExceptionUnchecked If there was an error with copying the jar file.
     */
    public File copyPluginJar(File destinationDir, String name) throws IOExceptionUnchecked
    {
        if (tester == null)
        {
            throw new IllegalStateException("Cannot copy the jar file before it has been created.");
        }
        final File destFile;
        if (name == null)
        {
            destFile = new File(destinationDir, tester.getPluginArchive().getName());
        } else
        {
            destFile = new File(destinationDir, name);
        }
        try
        {
            FileUtils.copyFile(tester.getPluginArchive(), destFile);
            return destFile;
        } catch (IOException ex)
        {
            throw CheckedExceptionTunnel.wrapIfNecessary(ex);
        }
    }

    private void checkTesterNull()
    {
        if (tester != null)
        {
            throw new IllegalStateException(
                    "Cannot change compilation units after plugin has been created.");
        }
    }

    private PluginTester tryCreatePluginTester(final File pluginArchive)
    {
        if (pluginArchive == null)
        {
            compiler.getLogWriter().write("Error creating jar file.");
            return null;
        }
        try
        {
            tester = new PluginTester(pluginArchive, compiler.getLogWriter(), requirePluginInfo);
        } catch (IllegalArgumentException ex)
        {
            compiler.getLogWriter().write(
                    "Error loading plugin: jar file does not contain a valid plugin.");
        }
        return tester;
    }

    @Override
    public void close() throws IOExceptionUnchecked
    {
        if (tester != null)
        {
            tester.close();
        }
        compiler.close();
    }
}
